@@ -1,6 +1,6 @@
 /*
 ** pork.h - pork main include file.
-** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2002-2004 Ryan McCabe <ryan@numb.org>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -18,8 +18,11 @@
 	#define __format(x)
 #endif
 
-#define AUTHOR		"Ryan McCabe <ryan@numb.org>"
-#define HOMEPAGE	"http://dev.ojnk.net"
+#ifdef ENABLE_DEBUGGING
+#	define debug(format, args...) do { fprintf(stderr, "[%s:%u:%s] DEBUG: " format "\n", __FILE__, __LINE__, __FUNCTION__, ##args); } while (0)
+#else
+#	define debug(format, args...) do { } while (0)
+#endif
 
 void pork_exit(int status, char *msg, char *fmt, ...) __format((printf, 3, 4));
 void keyboard_input(int fd, u_int32_t condition, void *data);
