@@ -1,6 +1,6 @@
 /*
 ** pork_aim.h - pork's interface with libfaim
-** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2002-2004 Ryan McCabe <ryan@numb.org>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -13,8 +13,7 @@
 #define AIM_DEFAULT_CHAT_EXCHANGE	4
 #define AIM_XFER_IN_HANDLER			1
 
-#define DEFAULT_AIM_PROFILE "<HTML><a href=\"http://dev.ojnk.net\">pork</a> - encore plus fort."
-
+#define DEFAULT_AIM_PROFILE "<HTML>i &lt;3 <a href=\"http://dev.ojnk.net\">pork</a>"
 #define FAIM_CB(x) static int x(aim_session_t *session __notused, aim_frame_t *fr __notused, ...)
 
 struct buddy;
@@ -40,7 +39,6 @@ struct chatroom_info {
 
 struct aim_priv {
 	u_int32_t marked_idle:1;
-	u_int32_t whois_hack:1;
 	time_t last_update;
 
 	aim_session_t aim_session;
