@@ -1,7 +1,7 @@
 /*
 ** pork_command.c - interface to commands typed by the user
-** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
-** Copyright (C) 2002-2003 Amber Adams <amber@ojnk.net>
+** Copyright (C) 2002-2004 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2002-2004 Amber Adams <amber@ojnk.net>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -52,9 +52,8 @@
 #include <pork_command_defs.h>
 #include <pork_help.h>
 
-#ifdef IRC_SUPPORT
-#	include <irc/pork_irc.h>
-#endif
+extern struct sockaddr_storage local_addr;
+extern in_port_t local_port;
 
 static void print_binding(void *data, void *nothing);
 static void print_alias(void *data, void *nothing);
@@ -75,30 +74,35 @@ static const struct command command[] = {
 	{ "",			cmd_send			},
 	{ "acct",		cmd_acct			},
 	{ "alias",		cmd_alias			},
+	{ "auto",		cmd_auto			},
 	{ "away",		cmd_away			},
 	{ "bind",		cmd_bind			},
 	{ "blist",		cmd_blist			},
 	{ "buddy",		cmd_buddy			},
 	{ "chat",		cmd_chat			},
 	{ "connect",	cmd_connect			},
+	{ "ctcp",		cmd_ctcp			},
 	{ "disconnect", cmd_disconnect		},
 	{ "echo",		cmd_echo			},
-	{ "email",		cmd_email			},
 	{ "event",		cmd_event			},
 	{ "file",		cmd_file			},
 	{ "help",		cmd_help			},
 	{ "history",	cmd_history			},
 	{ "idle",		cmd_idle			},
 	{ "input",		cmd_input			},
+	{ "laddr",		cmd_laddr			},
 	{ "lastlog",	cmd_lastlog			},
 	{ "load",		cmd_load			},
+	{ "lport",		cmd_lport			},
+	{ "me",			cmd_me				},
+	{ "mode",		cmd_mode			},
 	{ "msg",		cmd_msg				},
 	{ "nick",		cmd_nick			},
 	{ "notice",		cmd_notice			},
-	{ "passwd",		cmd_passwd			},
 	{ "perl",		cmd_perl			},
 	{ "perl_dump",	cmd_perl_dump		},
 	{ "perl_load",	cmd_perl_load		},
+	{ "ping",		cmd_ping			},
 	{ "profile",	cmd_profile,		},
 	{ "query",		cmd_query			},
 	{ "quit",		cmd_quit			},
@@ -128,6 +132,7 @@ static const struct command input_command[] = {
 	{ "clear_to_start",			cmd_input_clear_to_start	},
 	{ "delete",					cmd_input_delete			},
 	{ "end",					cmd_input_end				},
+	{ "find_next_cmd",			cmd_input_find_next_cmd		},
 	{ "focus_next",				cmd_input_focus_next		},
 	{ "insert",					cmd_input_insert			},
 	{ "left",					cmd_input_left				},
@@ -228,15 +233,15 @@ USER_COMMAND(cmd_input_send) {
 
 	recursion = 1;
 
-	if (args != NULL) {
-		xstrncpy(input->input_buf, args, sizeof(input->input_buf));
-		input->len = strlen(input->input_buf);
-	}
+	if (args != NULL)
+		input_set_buf(input, args);
 
 	if (input->len > 0) {
-		char *input_str = xstrdup(input->input_buf);
+		char *input_str = xstrdup(input_get_buf_str(input));
+
+		if (args == NULL)
+			input_history_add(input);
 
-		input_history_add(input);
 		input_clear_line(input);
 
 		if (event_generate(acct->events, EVENT_SEND_LINE,
@@ -313,14 +318,16 @@ USER_COMMAND(cmd_scroll_up) {
 }
 
 /*
-** file commands.
+** /file commands.
 */
 
 static const struct command file_command[] = {
+	{ "abort",				cmd_file_cancel		},
 	{ "cancel",				cmd_file_cancel 	},
-	{ "get",				cmd_file_get	 	},
-	{ "list",				cmd_file_list	 	},
-	{ "send",				cmd_file_send	 	},
+	{ "get",				cmd_file_get		},
+	{ "list",				cmd_file_list		},
+	{ "resume",				cmd_file_resume		},
+	{ "send",				cmd_file_send		},
 };
 
 USER_COMMAND(cmd_file_cancel) {
@@ -342,15 +349,12 @@ USER_COMMAND(cmd_file_cancel) {
 		return;
 	}
 
-	if (transfer_abort(xfer) != 0)
-		screen_err_msg("Error canceling file transfer %u", refnum);
-	else
-		screen_output("File transfer %u has been cancelled", refnum);
+	if (transfer_cancel_local(xfer) != 0)
+		screen_err_msg("Error canceling file transfer %s", args);
 }
 
 USER_COMMAND(cmd_file_list) {
 	struct pork_acct *acct = cur_window()->owner;
-	static const char *file_status[] = { "WAITING", "ACTIVE" };
 
 	if (acct->transfer_list != NULL) {
 		dlist_t *cur = acct->transfer_list;
@@ -361,7 +365,7 @@ USER_COMMAND(cmd_file_list) {
 			screen_output("%u: %s %s [%s: %u/%u (%.02f%%) - %.04f KB/s]",
 				xfer->refnum,
 				xfer->peer_username, xfer->fname_local,
-				file_status[xfer->status],
+				transfer_status_str(xfer),
 				(u_int32_t) xfer->bytes_sent, (u_int32_t) xfer->file_len,
 				(float) xfer->bytes_sent / xfer->file_len * 100,
 				transfer_avg_speed(xfer));
@@ -401,6 +405,9 @@ USER_COMMAND(cmd_file_get) {
 	transfer_get(xfer, args);
 }
 
+USER_COMMAND(cmd_file_resume) {
+}
+
 USER_COMMAND(cmd_file_send) {
 	char *dest;
 
@@ -417,13 +424,15 @@ USER_COMMAND(cmd_file_send) {
 }
 
 /*
-** win commands.
+** /win commands.
 */
 
 static const struct command window_command[] = {
 	{ "bind",				cmd_win_bind		},
+	{ "bind_next",			cmd_win_bind_next	},
 	{ "clear",				cmd_win_clear		},
 	{ "close",				cmd_win_close		},
+	{ "dump",				cmd_win_dump		},
 	{ "erase",				cmd_win_erase		},
 	{ "ignore",				cmd_win_ignore		},
 	{ "list",				cmd_win_list		},
@@ -460,7 +469,7 @@ USER_COMMAND(cmd_win_bind) {
 
 	ret = imwindow_bind_acct(imwindow, refnum);
 	if (ret == -1) {
-		if (imwindow->type == TYPE_CHAT)
+		if (imwindow->type == WIN_TYPE_CHAT)
 			screen_err_msg("You can't rebind chat windows");
 		else
 			screen_err_msg("Account %s isn't signed on", args);
@@ -470,6 +479,11 @@ USER_COMMAND(cmd_win_bind) {
 	}
 }
 
+USER_COMMAND(cmd_win_bind_next) {
+	if (imwindow_bind_next_acct(cur_window()) != -1)
+		screen_refresh();
+}
+
 USER_COMMAND(cmd_win_clear) {
 	imwindow_clear(cur_window());
 }
@@ -478,6 +492,16 @@ USER_COMMAND(cmd_win_close) {
 	screen_close_window(cur_window());
 }
 
+USER_COMMAND(cmd_win_dump) {
+	char buf[4096];
+
+	if (args == NULL || blank_str(args))
+		return;
+
+	expand_path(args, buf, sizeof(buf));
+	imwindow_dump_buffer(cur_window(), buf);
+}
+
 USER_COMMAND(cmd_win_erase) {
 	imwindow_erase(cur_window());
 }
@@ -844,7 +868,7 @@ USER_COMMAND(cmd_buddy_awaymsg) {
 		return;
 
 	if (args == NULL || blank_str(args)) {
-		if (win->type == TYPE_PRIVMSG)
+		if (win->type == WIN_TYPE_PRIVMSG)
 			args = win->target;
 		else
 			args = acct->username;
@@ -872,7 +896,7 @@ USER_COMMAND(cmd_buddy_privacy_mode) {
 		str_to_int(args, &mode);
 
 	mode = acct->proto->set_privacy_mode(acct, mode);
-	screen_output("privacy mode for %s is %d", acct->username, mode);
+	screen_output("Privacy mode for %s is %d", acct->username, mode);
 }
 
 USER_COMMAND(cmd_buddy_list) {
@@ -990,7 +1014,7 @@ USER_COMMAND(cmd_buddy_profile) {
 		return;
 
 	if (args == NULL || blank_str(args)) {
-		if (win->type == TYPE_PRIVMSG)
+		if (win->type == WIN_TYPE_PRIVMSG)
 			args = win->target;
 		else
 			args = acct->username;
@@ -1030,6 +1054,9 @@ USER_COMMAND(cmd_buddy_unblock) {
 USER_COMMAND(cmd_buddy_remove) {
 	struct pork_acct *acct = cur_window()->owner;
 
+	if (args == NULL || blank_str(args))
+		return;
+
 	if (buddy_remove(acct, args, 1) != 0)
 		screen_err_msg("%s isn't on %s's buddy list", args, acct->username);
 	else {
@@ -1123,7 +1150,7 @@ USER_COMMAND(cmd_buddy_warn) {
 		return;
 
 	if (args == NULL || blank_str(args)) {
-		if (win->type == TYPE_PRIVMSG)
+		if (win->type == WIN_TYPE_PRIVMSG)
 			args = win->target;
 		else
 			return;
@@ -1140,7 +1167,7 @@ USER_COMMAND(cmd_buddy_warn_anon) {
 		return;
 
 	if (args == NULL || blank_str(args)) {
-		if (win->type == TYPE_PRIVMSG)
+		if (win->type == WIN_TYPE_PRIVMSG)
 			args = win->target;
 		else
 			return;
@@ -1172,6 +1199,7 @@ static const struct command blist_command[] = {
 	{ "select",			cmd_blist_select		},
 	{ "show",			cmd_blist_show			},
 	{ "start",			cmd_blist_start			},
+	{ "toggle",			cmd_blist_toggle		},
 	{ "up",				cmd_blist_up			},
 	{ "warn",			cmd_blist_warn			},
 	{ "warn_anon",		cmd_blist_warn_anon		},
@@ -1364,6 +1392,10 @@ USER_COMMAND(cmd_blist_start) {
 	blist_cursor_start(cur_window()->owner->blist);
 }
 
+USER_COMMAND(cmd_blist_toggle) {
+	imwindow_blist_toggle(cur_window());
+}
+
 USER_COMMAND(cmd_blist_up) {
 	blist_cursor_up(cur_window()->owner->blist);
 }
@@ -1409,6 +1441,10 @@ USER_COMMAND(cmd_blist_width) {
 	screen_blist_width(blist, new_len);
 }
 
+/*
+** /timer commands
+*/
+
 static const struct command timer_command[] = {
 	{ "add",			cmd_timer_add			},
 	{ "del",			cmd_timer_del			},
@@ -1587,22 +1623,51 @@ USER_COMMAND(cmd_event_purge) {
 }
 
 /*
-** Chat commands
+** /chat commands
 */
 
 static const struct command chat_command[] = {
+	{ "ban",				cmd_chat_ban			},
 	{ "ignore",				cmd_chat_ignore			},
 	{ "invite",				cmd_chat_invite			},
 	{ "join",				cmd_chat_join			},
 	{ "kick",				cmd_chat_kick			},
 	{ "leave",				cmd_chat_leave			},
 	{ "list",				cmd_chat_list			},
+	{ "mode",				cmd_chat_mode			},
 	{ "send",				cmd_chat_send			},
 	{ "topic",				cmd_chat_topic			},
 	{ "unignore",			cmd_chat_unignore		},
 	{ "who",				cmd_chat_who			},
 };
 
+USER_COMMAND(cmd_chat_ban) {
+	struct imwindow *win = cur_window();
+	struct pork_acct *acct = win->owner;
+	struct chatroom *chat;
+	char *arg1;
+	char *arg2;
+
+	if (args == NULL)
+		return;
+
+	arg1 = strsep(&args, " ");
+
+	chat = chat_find(acct, arg1);
+	if (chat == NULL) {
+		if (win->type == WIN_TYPE_CHAT && win->data != NULL)
+			chat_ban(acct, win->data, arg1);
+		else
+			screen_err_msg("%s is not a member of %s", acct->username, arg1);
+
+		return;
+	}
+
+	arg2 = strsep(&args, " ");
+	if (arg2 != NULL)
+		chat_ban(acct, chat, arg2);
+}
+
 USER_COMMAND(cmd_chat_ignore) {
 	struct imwindow *imwindow = cur_window();
 	struct pork_acct *acct = imwindow->owner;
@@ -1618,7 +1683,7 @@ USER_COMMAND(cmd_chat_ignore) {
 	if (user_name == NULL) {
 		struct chatroom *chat = imwindow->data;
 
-		if (imwindow->type != TYPE_CHAT || chat == NULL) {
+		if (imwindow->type != WIN_TYPE_CHAT || chat == NULL) {
 			screen_err_msg("You must specify a chat room if the current window is not a chat window");
 			return;
 		}
@@ -1647,7 +1712,7 @@ USER_COMMAND(cmd_chat_invite) {
 	if (user_name == NULL) {
 		struct chatroom *chat = imwindow->data;
 
-		if (imwindow->type != TYPE_CHAT || chat == NULL) {
+		if (imwindow->type != WIN_TYPE_CHAT || chat == NULL) {
 			screen_err_msg("You must specify a chat room if the current window is not a chat window");
 			return;
 		}
@@ -1677,7 +1742,7 @@ USER_COMMAND(cmd_chat_kick) {
 
 	chat = chat_find(acct, arg1);
 	if (chat == NULL) {
-		if (win->type == TYPE_CHAT && win->data != NULL)
+		if (win->type == WIN_TYPE_CHAT && win->data != NULL)
 			chat_kick(acct, win->data, arg1, args);
 		else
 			screen_err_msg("%s is not a member of %s", acct->username, arg1);
@@ -1697,7 +1762,7 @@ USER_COMMAND(cmd_chat_leave) {
 	if (name == NULL || blank_str(name)) {
 		struct chatroom *chat;
 
-		if (win->type != TYPE_CHAT) {
+		if (win->type != WIN_TYPE_CHAT) {
 			screen_err_msg("You must specify a chat room if the current window is not a chat window");
 			return;
 		}
@@ -1716,6 +1781,10 @@ USER_COMMAND(cmd_chat_list) {
 	chat_list(cur_window()->owner);
 }
 
+USER_COMMAND(cmd_chat_mode) {
+	/* XXX: TODO */
+}
+
 USER_COMMAND(cmd_chat_send) {
 	struct pork_acct *acct = cur_window()->owner;
 	struct imwindow *win;
@@ -1736,7 +1805,7 @@ USER_COMMAND(cmd_chat_send) {
 		return;
 	}
 
-	chat_send_msg(acct, win->data, args);
+	chat_send_msg(acct, win->data, chat_name, args);
 }
 
 USER_COMMAND(cmd_chat_topic) {
@@ -1745,7 +1814,7 @@ USER_COMMAND(cmd_chat_topic) {
 	char *topic = NULL;
 	struct chatroom *chat = NULL;
 
-	if (acct->proto->chat_set_topic == NULL || acct->proto->is_chat == NULL)
+	if (acct->proto->chat_set_topic == NULL)
 		return;
 
 	if (args != NULL) {
@@ -1762,9 +1831,9 @@ USER_COMMAND(cmd_chat_topic) {
 
 		topic = args;
 
-		if (win->type == TYPE_CHAT)
+		if (win->type == WIN_TYPE_CHAT)
 			chat = win->data;
-		 else {
+		else {
 			screen_err_msg("You must specify a chat room if the current window isn't a chat window");
 			return;
 		}
@@ -1788,7 +1857,7 @@ USER_COMMAND(cmd_chat_unignore) {
 	if (user_name == NULL) {
 		struct chatroom *chat = imwindow->data;
 
-		if (imwindow->type != TYPE_CHAT || chat == NULL) {
+		if (imwindow->type != WIN_TYPE_CHAT || chat == NULL) {
 			screen_err_msg("You must specify a chat room if the current window is not a chat window");
 			return;
 		}
@@ -1807,7 +1876,7 @@ USER_COMMAND(cmd_chat_who) {
 	if (args == NULL || blank_str(args)) {
 		struct chatroom *chat;
 
-		if (imwindow->type != TYPE_CHAT) {
+		if (imwindow->type != WIN_TYPE_CHAT) {
 			screen_err_msg("You must specify a chat room if the current window is not a chat window");
 			return;
 		}
@@ -1816,8 +1885,7 @@ USER_COMMAND(cmd_chat_who) {
 		args = chat->title;
 	}
 
-	if (chat_who(acct, args) != 0)
-		screen_err_msg("%s is not a member of %s", acct->username, args);
+	chat_who(acct, args);
 }
 
 static const struct {
@@ -1826,7 +1894,7 @@ static const struct {
 	char *type;
 } command_set[] = {
 	{	command,			array_elem(command),			"" 			},
-	{	window_command,		array_elem(window_command),		"window "	},
+	{	window_command,		array_elem(window_command),		"win "		},
 	{	history_command,	array_elem(history_command),	"history "	},
 	{	input_command,		array_elem(input_command),		"input "	},
 	{	scroll_command,		array_elem(scroll_command),		"scroll "	},
@@ -1885,6 +1953,20 @@ USER_COMMAND(cmd_alias) {
 	screen_err_msg("Error adding alias for %s", alias);
 }
 
+USER_COMMAND(cmd_auto) {
+	struct pork_acct *acct = cur_window()->owner;
+	char *target;
+
+	if (args == NULL || !acct->connected)
+		return;
+
+	target = strsep(&args, " ");
+	if (target == NULL || args == NULL)
+		return;
+
+	pork_msg_autoreply(acct, target, args);
+}
+
 USER_COMMAND(cmd_away) {
 	struct pork_acct *acct = cur_window()->owner;
 
@@ -1960,74 +2042,43 @@ USER_COMMAND(cmd_bind) {
 }
 
 USER_COMMAND(cmd_connect) {
-	struct pork_acct *acct = NULL;
-	dlist_t *node;
 	int protocol = PROTO_AIM;
+	char *user;
 
-	if (args != NULL && !blank_str(args)) {
-		char *user = args;
+	if (args == NULL || blank_str(args))
+		return;
 
-#ifdef IRC_SUPPORT
-		if (!strncasecmp(args, "-irc ", 5)) {
-			args += 5;
-			protocol = PROTO_IRC;
-			user = strsep(&args, " ");
-		}
-#endif
+	if (*args == '-') {
+		char *p = strchr(++args, ' ');
 
-		/* XXX - check for protocol when others are added */
-		acct = pork_acct_init(user, protocol);
-		if (acct == NULL) {
-			screen_err_msg("%s is already connected", user);
+		if (p != NULL)
+			*p++ = '\0';
+
+		protocol = proto_get_num(args);
+		if (protocol == -1) {
+			screen_err_msg("Invalid protocol: %s", args);
 			return;
 		}
 
-		pork_acct_add(acct);
-
-		node = screen.window_list;
-
-		do {
-			struct imwindow *imwindow = node->data;
-
-			if (imwindow->owner == screen.null_acct) {
-				imwindow_bind_acct(imwindow, acct->refnum);
-
-				if (!imwindow->blist_visible) {
-					if (wopt_get_bool(imwindow->opts, WOPT_SHOW_BLIST))
-						wopt_set(imwindow, WOPT_SHOW_BLIST, "1");
-				}
-			}
-
-			node = node->next;
-		} while (node != screen.window_list);
-	} else {
-		struct imwindow *imwindow = cur_window();
-
-		if (imwindow->owner->connected) {
-			char buf[32];
-
-			snprintf(buf, sizeof(buf), "%u", imwindow->owner->refnum);
-			cmd_win_bind(buf);
-			return;
-		}
-
-		acct = imwindow->owner;
+		args = p;
 	}
 
-	if (!acct->can_connect) {
-		screen_err_msg("You must specify a screen name before connecting");
-		return;
-	}
+	user = strsep(&args, " ");
+	pork_acct_connect(user, args, protocol);
+}
 
-	if (acct->proto->connect == NULL)
+USER_COMMAND(cmd_ctcp) {
+	struct pork_acct *acct = cur_window()->owner;
+	char *dest;
+
+	if (acct->proto->ctcp == NULL)
 		return;
 
-	if (acct->proto->connect(acct, args) == -1) {
-		screen_err_msg("Unable to login as %s", acct->username);
-		pork_acct_del_refnum(acct->refnum, NULL);
-	}
+	dest = strsep(&args, " ");
+	if (dest == NULL || args == NULL)
+		return;
 
-	return;
+	acct->proto->ctcp(acct, dest, args);
 }
 
 USER_COMMAND(cmd_echo) {
@@ -2040,7 +2091,7 @@ USER_COMMAND(cmd_disconnect) {
 	u_int32_t dest;
 	dlist_t *node;
 
-	if (!acct->connected)
+	if (!acct->can_connect)
 		return;
 
 	if (args == NULL || blank_str(args))
@@ -2070,6 +2121,9 @@ USER_COMMAND(cmd_disconnect) {
 	}
 
 	pork_acct_del(node, args);
+
+	if (screen.status_win->owner == screen.null_acct)
+		imwindow_bind_next_acct(screen.status_win);
 }
 
 USER_COMMAND(cmd_help) {
@@ -2132,6 +2186,24 @@ USER_COMMAND(cmd_idle) {
 	pork_set_idle_time(cur_window()->owner, idle_secs);
 }
 
+USER_COMMAND(cmd_laddr) {
+	if (args == NULL) {
+		char buf[2048];
+
+		if (get_hostname(&local_addr, buf, sizeof(buf)) != 0)
+			xstrncpy(buf, "0.0.0.0", sizeof(buf));
+
+		screen_output("New connections will use the local address %s", buf);
+		return;
+	}
+
+	if (get_addr(args, &local_addr) != 0) {
+		screen_err_msg("Invalid local address: %s", args);
+		return;
+	}
+
+	screen_output("New connections will use the local address %s", args);
+}
 
 USER_COMMAND(cmd_lastlog) {
 	int opts = 0;
@@ -2188,6 +2260,38 @@ USER_COMMAND(cmd_load) {
 	screen_set_quiet(quiet);
 }
 
+USER_COMMAND(cmd_lport) {
+	if (args == NULL) {
+		screen_output("New connections will use local port %u", ntohs(local_port));
+		return;
+	}
+
+	if (get_port(args, &local_port) != 0) {
+		screen_err_msg("Error: Invalid local port: %s", args);
+		return;
+	}
+
+	local_port = htons(local_port);
+	screen_output("New connections will use local port %s", args);
+}
+
+USER_COMMAND(cmd_me) {
+	struct imwindow *win = cur_window();
+
+	if (args == NULL)
+		return;
+
+	if (win->type == WIN_TYPE_PRIVMSG)
+		pork_action_send(win->owner, cur_window()->target, args);
+	else if (win->type == WIN_TYPE_CHAT) {
+		struct chatroom *chat;
+
+		chat = win->data;
+		if (chat != NULL)
+			chat_send_action(win->owner, win->data, chat->title, args);
+	}
+}
+
 USER_COMMAND(cmd_msg) {
 	struct pork_acct *acct = cur_window()->owner;
 	char *target;
@@ -2202,37 +2306,35 @@ USER_COMMAND(cmd_msg) {
 
 	chat = chat_find(acct, target);
 	if (chat != NULL)
-		chat_send_msg(acct, chat, args);
+		chat_send_msg(acct, chat, target, args);
 	else
 		pork_msg_send(acct, target, args);
 }
 
+USER_COMMAND(cmd_mode) {
+	struct pork_acct *acct = cur_window()->owner;
+
+	if (args == NULL || !acct->connected)
+		return;
+
+	if (acct->proto->mode != NULL)
+		acct->proto->mode(acct, args);
+}
+
 USER_COMMAND(cmd_query) {
 	struct imwindow *imwindow = cur_window();
 
 	if (args != NULL && !blank_str(args)) {
 		struct imwindow *conv_window;
 
-		conv_window = imwindow_find(imwindow->owner, args);
-		if (conv_window == NULL) {
-			struct buddy *buddy = buddy_find(imwindow->owner, args);
-			char *name;
-
-			if (buddy != NULL)
-				name = buddy->name;
-			else
-				name = args;
-
-			conv_window = screen_new_window(imwindow->owner, args, name);
-		}
-
+		screen_make_query_window(imwindow->owner, args, &conv_window);
 		screen_goto_window(conv_window->refnum);
 	} else
 		screen_close_window(imwindow);
 }
 
 USER_COMMAND(cmd_quit) {
-	if (!event_generate(cur_window()->owner->events, EVENT_QUIT))
+	if (!event_generate(cur_window()->owner->events, EVENT_QUIT, args))
 		pork_exit(0, args, NULL);
 }
 
@@ -2263,16 +2365,16 @@ USER_COMMAND(cmd_send) {
 	if (args == NULL || !acct->connected)
 		return;
 
-	if (imwindow->type == TYPE_PRIVMSG)
+	if (imwindow->type == WIN_TYPE_PRIVMSG)
 		pork_msg_send(acct, imwindow->target, args);
-	else if (imwindow->type == TYPE_CHAT) {
+	else if (imwindow->type == WIN_TYPE_CHAT) {
 		struct chatroom *chat = imwindow->data;
 
 		if (chat == NULL) {
 			screen_err_msg("%s is not a member of %s",
 				acct->username, imwindow->target);
 		} else
-			chat_send_msg(acct, chat, args);
+			chat_send_msg(acct, chat, chat->title, args);
 	}
 }
 
@@ -2324,9 +2426,6 @@ USER_COMMAND(cmd_unalias) {
 		screen_status_msg("Alias %s removed", args);
 }
 
-USER_COMMAND(cmd_email) {
-}
-
 USER_COMMAND(cmd_nick) {
 	if (args == NULL || blank_str(args))
 		return;
@@ -2348,19 +2447,25 @@ USER_COMMAND(cmd_notice) {
 
 	chat = chat_find(acct, target);
 	if (chat != NULL)
-		chat_send_notice(acct, chat, args);
+		chat_send_notice(acct, chat, target, args);
 	else
 		pork_notice_send(acct, target, args);
 }
 
-USER_COMMAND(cmd_passwd) {
-}
-
 USER_COMMAND(cmd_who) {
-	struct pork_acct *acct = cur_window()->owner;
+	struct imwindow *win = cur_window();
+	struct pork_acct *acct = win->owner;
 
 	if (acct->proto->who != NULL && args != NULL)
 		acct->proto->who(acct, args);
+
+	if (args == NULL && win->type == WIN_TYPE_CHAT) {
+		if (win->data != NULL) {
+			struct chatroom *chat = win->data;
+
+			chat_who(acct, chat->title);
+		}
+	}
 }
 
 USER_COMMAND(cmd_whowas) {
@@ -2475,6 +2580,14 @@ USER_COMMAND(cmd_perl_load) {
 		screen_err_msg("Error: The file %s couldn't be loaded", buf);
 }
 
+USER_COMMAND(cmd_ping) {
+	struct imwindow *win = cur_window();
+	struct pork_acct *acct = win->owner;
+
+	if (acct->proto->ping != NULL)
+		acct->proto->ping(acct, args);
+}
+
 USER_COMMAND(cmd_profile) {
 	pork_set_profile(cur_window()->owner, args);
 }
@@ -2681,3 +2794,153 @@ static void print_timer(void *data, void *nothing __notused) {
 	screen_output("[refnum: %u] %d %u %s", timer->refnum,
 		(int) timer->interval, timer->times, timer->command);
 }
+
+USER_COMMAND(cmd_input_find_next_cmd) {
+	struct input *input = cur_window()->input;
+	u_int16_t cur_pos = input->cur - input->prompt_len;
+	u_int16_t begin_completion_pos;
+	char *input_buf = input_get_buf_str(input);
+
+	if (input->begin_completion <= input->prompt_len ||
+		input->begin_completion >= input->cur)
+	{
+		if (cur_pos == 0) {
+			/*
+			** If you complete from the very beginning of the line,
+			** insert a '/' because we only complete commands!
+			*/
+
+			input_insert(input, '/');
+			input_buf = input_get_buf_str(input);
+			cur_pos++;
+		} else
+			input->begin_completion = input->cur;
+	}
+
+	begin_completion_pos = input->begin_completion - input->prompt_len;
+
+	/* Only complete if the line is a command. */
+	if (input_buf[0] == '/') {
+		/* find first word... */
+		u_int32_t end_word = strcspn(&input_buf[1], " \t") + 1;
+		u_int32_t word_begin = 0;
+		u_int16_t cmdset = 0;
+
+		if (end_word < begin_completion_pos) {
+			/* checking second word! */
+			size_t i;
+
+			for (i = 1 ; i < array_elem(command_set) ; i++) {
+				if (!strncasecmp(command_set[i].type, &input_buf[1], end_word))
+					cmdset = i;
+			}
+
+			if (cmdset != 0) {
+				word_begin = end_word;
+				while (	input_buf[word_begin] == ' ' ||
+						input_buf[word_begin] == '\t')
+				{
+					word_begin++;
+				}
+			}
+		} else {
+			/* checking first word */
+			word_begin = 1;
+			cmdset = CMDSET_MAIN;
+		}
+
+		if (word_begin > 0) {
+			int chosen = -1;
+			size_t i;
+
+			for (i = 0 ; i < command_set[cmdset].elem ; i++) {
+				const struct command *cmd_ptr;
+
+				cmd_ptr = command_set[cmdset].set + i;
+				if (!strcmp(cmd_ptr->name, ""))
+					continue;
+
+				if (input->cur == input->begin_completion) {
+					if (!strncasecmp(cmd_ptr->name,
+							&input_buf[word_begin], cur_pos - word_begin))
+					{
+						/*
+						** Don't 'choose' this one if
+						** it's already an exact match!
+						*/
+
+						if (strlen(cmd_ptr->name) == cur_pos - word_begin)
+							continue;
+
+						/* MATCH! */
+						chosen = i;
+						i = command_set[cmdset].elem;
+						continue;
+					}
+				} else {
+					/*
+					** We've _already_ matched one. Now get
+					** the next in the list, if possible.
+					*/
+
+					if (!strncasecmp(cmd_ptr->name,
+							&input_buf[word_begin], cur_pos - word_begin))
+					{
+						/*
+						** We've found the one we've matched already. If there
+						** will be another possible, it will be the very next
+						** one in the list. Let's check that. If it doesn't
+						** match, none will.
+						*/
+
+						i++;
+						if (i < command_set[cmdset].elem) {
+							cmd_ptr = command_set[cmdset].set + i;
+
+							if (!strncasecmp(cmd_ptr->name,
+									&input_buf[word_begin],
+									begin_completion_pos - word_begin))
+							{
+								/*
+								** The next one matched!
+								** Let's pick it!
+								*/
+
+								chosen = i;
+							}
+
+							i = command_set[cmdset].elem;
+							continue;
+						}
+					}
+				}
+			}
+
+			if (input->begin_completion != input->cur) {
+				/*
+				** Remove characters added by last match. In
+				** other words, remove characters from input->cur
+				** until input->begin_completion == input->cur.
+				*/
+				size_t num_to_remove;
+
+				num_to_remove = input->cur - input->begin_completion;
+
+				for (i = 0 ; i < num_to_remove ; i++)
+					input_bkspace(input);
+			}
+
+			if (chosen != -1) {
+				/* We've found a match. */
+				const struct command *cmd_ptr;
+				u_int16_t remember_begin = input->begin_completion;
+
+				cmd_ptr = command_set[cmdset].set + chosen;
+
+				input_insert_str(input,
+					&cmd_ptr->name[begin_completion_pos - word_begin]);
+				input->begin_completion = remember_begin;
+			}
+		}
+	}
+}
