@@ -1,7 +1,7 @@
 /*
 ** pork_format.c - helper functions for formatting text.
-** Copyright (C) 2002-2003 Amber Adams <amber@ojnk.net>
-** Copyright (C) 2003 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2002-2004 Amber Adams <amber@ojnk.net>
+** Copyright (C) 2003-2004 Ryan McCabe <ryan@numb.org>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -17,6 +17,7 @@
 #include <time.h>
 #include <ctype.h>
 #include <time.h>
+#include <errno.h>
 #include <sys/time.h>
 #include <sys/types.h>
 
@@ -29,11 +30,13 @@
 #include <pork_imwindow.h>
 #include <pork_buddy_list.h>
 #include <pork_acct.h>
+#include <pork_proto.h>
 #include <pork_cstr.h>
 #include <pork_misc.h>
 #include <pork_screen.h>
 #include <pork_screen_cmd.h>
 #include <pork_transfer.h>
+#include <pork_chat.h>
 #include <pork_format.h>
 
 static int format_status_timestamp(	char opt,
@@ -41,15 +44,17 @@ static int format_status_timestamp(	char opt,
 									size_t len,
 									va_list ap __notused)
 {
-	int ret = 0;
 	time_t cur_time;
 	struct tm *tm;
+	int ret = 0;
 
 	cur_time = time(NULL);
 	tm = localtime(&cur_time);
 
-	if (tm == NULL)
+	if (tm == NULL) {
+		debug("localtime: %s", strerror(errno));
 		return (-1);
+	}
 
 	switch (opt) {
 		/* Hours, 24-hour format */
@@ -58,8 +63,7 @@ static int format_status_timestamp(	char opt,
 			break;
 
 		/* Hours, 12-hour format */
-		case 'h':
-		{
+		case 'h': {
 			u_int32_t time_12hour;
 
 			if (tm->tm_hour == 0)
@@ -109,8 +113,7 @@ static int format_status_activity(	char opt,
 	switch (opt) {
 		/* Activity display */
 		case 'A':
-		case 'a':
-		{
+		case 'a': {
 			dlist_t *cur = screen.window_list;
 			size_t n = 0;
 
@@ -180,7 +183,6 @@ static int format_status_typing(char opt, char *buf, size_t len, va_list ap) {
 static int format_status_held(char opt, char *buf, size_t len, va_list ap) {
 	struct imwindow *imwindow = va_arg(ap, struct imwindow *);
 	int ret = 0;
-	int r_code;
 
 	if (imwindow->swindow.held == 0)
 		return (1);
@@ -188,21 +190,23 @@ static int format_status_held(char opt, char *buf, size_t len, va_list ap) {
 	switch (opt) {
 		/* Held indicator */
 		case 'H':
-			r_code = snprintf(buf, len, "%02d", imwindow->swindow.held);
-			if (r_code < 0 || (size_t) r_code >= len)
-				ret = -1;
+			ret = snprintf(buf, len, "%02d", imwindow->swindow.held);
 			break;
 
 		default:
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_status_idle(char opt, char *buf, size_t len, va_list ap) {
 	struct pork_acct *acct = va_arg(ap, struct pork_acct *);
 	int hide_if_zero = va_arg(ap, int);
+	int ret = 0;
 
 	switch (opt) {
 		/* Idle time */
@@ -211,25 +215,24 @@ static int format_status_idle(char opt, char *buf, size_t len, va_list ap) {
 			if (acct->idle_time == 0 && hide_if_zero)
 				return (1);
 
-			time_to_str(acct->idle_time, buf, len);
+			ret = time_to_str(acct->idle_time, buf, len);
 			break;
 
 		/*
 		** If the current window is a chat window, $i will display the
 		** idle time of the user we're talking to.
 		*/
-		case 'i':
-		{
+		case 'i': {
 			struct imwindow *win = cur_window();
 
-			if (win->type == TYPE_PRIVMSG) {
+			if (win->type == WIN_TYPE_PRIVMSG) {
 				struct buddy *buddy;
 
 				buddy = buddy_find(acct, win->target);
 				if (buddy == NULL || (buddy->idle_time == 0 && hide_if_zero))
 					return (1);
 
-				time_to_str(buddy->idle_time, buf, len);
+				ret = time_to_str(buddy->idle_time, buf, len);
 				break;
 			}
 
@@ -240,6 +243,9 @@ static int format_status_idle(char opt, char *buf, size_t len, va_list ap) {
 			return (-1);
 	}
 
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
 	return (0);
 }
 
@@ -249,29 +255,115 @@ static int format_status(char opt, char *buf, size_t len, va_list ap) {
 	int ret = 0;
 
 	switch (opt) {
+		/* Disconnect / reconnect status */
+		case 'D':
+			if (acct->reconnecting)
+				ret = xstrncpy(buf, "reconnecting", len);
+			else if (acct->disconnected)
+				ret = xstrncpy(buf, "disconnected", len);
+			else
+				ret = xstrncpy(buf, "connected", len);
+			break;
+
 		/* Screen name */
 		case 'n':
 		case 'N':
-			if (xstrncpy(buf, acct->username, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, acct->username, len);
+			break;
+
+		/* Remote server */
+		case 'S':
+		case 's':
+			if (acct->server != NULL)
+				ret = xstrncpy(buf, acct->server, len);
+			break;
+
+		/* Remote port */
+		case 'P':
+		case 'p':
+			if (acct->fport != NULL)
+				ret = xstrncpy(buf, acct->fport, len);
 			break;
 
 		/* Window name */
 		case 'z':
 		case 'Z':
-			if (xstrncpy(buf, imwindow->name, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, imwindow->name, len);
 			break;
 
 		/* Chat room, if inactive */
 		case 'c':
 		case 'C':
-			if (imwindow->type == TYPE_CHAT && imwindow->data == NULL) {
-				if (xstrncpy(buf, opt_get_str(OPT_TEXT_NO_ROOM), len) == -1)
-					ret = -1;
+			if (imwindow->type == WIN_TYPE_CHAT &&
+				(imwindow->data == NULL || !acct->connected))
+			{
+				ret = xstrncpy(buf, opt_get_str(OPT_TEXT_NO_ROOM), len);
 			}
 			break;
 
+		/* Chat mode, if applicable; M includes arguments, m doesn't. */
+		case 'M':
+		case 'm':
+			if (imwindow->type == WIN_TYPE_CHAT && imwindow->data != NULL) {
+				struct chatroom *chat = imwindow->data;
+
+				ret = xstrncpy(buf, chat->mode, len);
+				if (opt == 'm') {
+					char *p;
+
+					p = strchr(buf, ' ');
+					if (p != NULL)
+						*p = '\0';
+				}
+			}
+			break;
+
+		/* Chat status, if applicable */
+		case '@':
+			if (imwindow->type == WIN_TYPE_CHAT && imwindow->data != NULL) {
+				struct chatroom *chat = imwindow->data;
+				struct chat_user *user;
+
+				user = chat_find_user(acct, chat, acct->username);
+				if (user == NULL)
+					break;
+
+				if (user->status & CHAT_STATUS_OP)
+					ret = xstrncpy(buf, "@", len);
+				else if (user->status & CHAT_STATUS_HALFOP)
+					ret = xstrncpy(buf, "%%", len);
+				else if (user->status & CHAT_STATUS_VOICE)
+					ret = xstrncpy(buf, "+", len);
+			}
+			break;
+
+		/* User status */
+		case '!':
+			if (acct->disconnected) {
+				if (acct->reconnecting)
+					ret = xstrncpy(buf, "reconnecting", len);
+				else
+					ret = xstrncpy(buf, "disconnected", len);
+			} else if (acct->connected) {
+				if (acct->away_msg != NULL)
+					ret = xstrncpy(buf, "away", len);
+				else
+					ret = xstrncpy(buf, "online", len);
+			} else
+				ret = xstrncpy(buf, "not connected", len);
+			break;
+
+		/* Protocol */
+		case '?':
+			ret = xstrncpy(buf, acct->proto->name, len);
+			break;
+
+		/* User mode */
+		case 'u':
+		case 'U':
+			ret = xstrncpy(buf, acct->umode, len);
+			break;
+
 		/* Timestamp */
 		case 't':
 		case 'T':
@@ -314,14 +406,16 @@ static int format_status(char opt, char *buf, size_t len, va_list ap) {
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_status_warn(char opt, char *buf, size_t len, va_list ap) {
 	struct pork_acct *acct= va_arg(ap, struct pork_acct *);
 	int hide_if_zero = va_arg(ap, int);
 	int ret = 0;
-	int r_code;
 
 	switch (opt) {
 		/* Warn level */
@@ -330,25 +424,20 @@ static int format_status_warn(char opt, char *buf, size_t len, va_list ap) {
 			if (acct->warn_level == 0 && hide_if_zero)
 				return (1);
 
-			r_code = snprintf(buf, len, "%02d", acct->warn_level);
-			if (r_code < 0 || (size_t) r_code >= len)
-				ret = -1;
+			ret = snprintf(buf, len, "%02d", acct->warn_level);
 			break;
 
-		case 'w':
-		{
+		case 'w': {
 			struct imwindow *win = cur_window();
 
-			if (win->type == TYPE_PRIVMSG) {
+			if (win->type == WIN_TYPE_PRIVMSG) {
 				struct buddy *buddy;
 
 				buddy = buddy_find(acct, win->target);
 				if (buddy == NULL || (buddy->warn_level == 0 && hide_if_zero))
 					return (1);
 
-				r_code = snprintf(buf, len, "%02d", buddy->warn_level);
-				if (r_code < 0 || (size_t) r_code >= len)
-					ret = -1;
+				ret = snprintf(buf, len, "%02d", buddy->warn_level);
 				break;
 			}
 
@@ -359,156 +448,350 @@ static int format_status_warn(char opt, char *buf, size_t len, va_list ap) {
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
-static int format_privmsg(char opt, char *buf, size_t len, va_list ap) {
-	char *nick = va_arg(ap, char *);
+static int format_msg_send(char opt, char *buf, size_t len, va_list ap) {
+	struct pork_acct *acct = va_arg(ap, struct pork_acct *);
 	char *dest = va_arg(ap, char *);
-	char *msg = va_arg(ap, char *);
+ 	char *msg = va_arg(ap, char *);
 	int ret = 0;
 
 	switch (opt) {
 		/* Screen name of sender */
 		case 'n':
 		case 'N':
-			if (xstrncpy(buf, nick, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, acct->username, len);
 			break;
 
-		/* Screen name of the receiver */
+		/* Screen name / alias  of the receiver */
 		case 'R':
+			if (dest != NULL)
+				ret = xstrncpy(buf, buddy_name(acct, dest), len);
+			break;
+
 		case 'r':
-			if (xstrncpy(buf, dest, len) == -1)
-				ret = -1;
+			if (dest != NULL)
+				ret = xstrncpy(buf, dest, len);
 			break;
 
 		/* Message text */
-		case 'm':
 		case 'M':
-			if (xstrncpy(buf, msg, len) == -1)
-				ret = -1;
+			if (msg != NULL) {
+				msg = acct->proto->filter_text(msg);
+				ret = xstrncpy(buf, msg, len);
+				free(msg);
+			}
+			break;
+
+		case 'm':
+			if (msg != NULL)
+				ret = xstrncpy(buf, msg, len);
+			break;
+
+		case 'H':
+			if (acct->userhost != NULL) {
+				char *host = acct->proto->filter_text(acct->userhost);
+				ret = xstrncpy(buf, host, len);
+				free(host);
+			}
 			break;
 
 		default:
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+
+	return (0);
 }
 
-static int format_chat_msg(char opt, char *buf, size_t len, va_list ap) {
-	char *source = va_arg(ap, char *);
+static int format_msg_recv(char opt, char *buf, size_t len, va_list ap) {
+	struct pork_acct *acct = va_arg(ap, struct pork_acct *);
+	char *dest = va_arg(ap, char *);
+ 	char *sender = va_arg(ap, char *);
+	char *sender_userhost = va_arg(ap, char *);
+	char *msg = va_arg(ap, char *);
+	int ret = 0;
+
+	switch (opt) {
+		/* Screen name of sender */
+		case 'n':
+			if (sender != NULL)
+				ret = xstrncpy(buf, sender, len);
+			break;
+
+		/* Screen name / alias of sender */
+		case 'N':
+			if (sender != NULL)
+				ret = xstrncpy(buf, buddy_name(acct, sender), len);
+			break;
+
+		/* Screen name / alias  of the receiver */
+		case 'R':
+			if (dest != NULL)
+				ret = xstrncpy(buf, buddy_name(acct, dest), len);
+			break;
+
+		case 'r':
+			if (dest != NULL)
+				ret = xstrncpy(buf, dest, len);
+			break;
+
+		/* Message text */
+		case 'm':
+		case 'M':
+			if (msg != NULL) {
+				msg = acct->proto->filter_text(msg);
+				ret = xstrncpy(buf, msg, len);
+				free(msg);
+			}
+			break;
+
+		case 'H':
+			if (acct->userhost != NULL) {
+				char *host = acct->proto->filter_text(acct->userhost);
+				ret = xstrncpy(buf, host, len);
+				free(host);
+			}
+			break;
+
+		case 'h':
+			if (sender_userhost != NULL) {
+				char *host = acct->proto->filter_text(sender_userhost);
+				ret = xstrncpy(buf, host, len);
+				free(host);
+			}
+			break;
+
+		default:
+			return (-1);
+	}
+
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
+}
+
+static int format_chat_send(char opt, char *buf, size_t len, va_list ap) {
+	struct pork_acct *acct = va_arg(ap, struct pork_acct *);
+	struct chatroom *chat = va_arg(ap, struct chatroom *);
 	char *dest = va_arg(ap, char *);
 	char *msg = va_arg(ap, char *);
+
+	int ret = 0;
+
+	switch (opt) {
+		/* Message source */
+		case 'n':
+		case 'N':
+			ret = xstrncpy(buf, acct->username, len);
+			break;
+
+		/* Message destination */
+		case 'C':
+			if (chat != NULL)
+				ret = xstrncpy(buf, chat->title_quoted, len);
+			break;
+
+		case 'c':
+			if (dest != NULL) {
+				dest = acct->proto->filter_text(dest);
+				ret = xstrncpy(buf, dest, len);
+				free(dest);
+			}
+			break;
+
+		/* Message text */
+		case 'm':
+		case 'M':
+			if (msg != NULL) {
+				msg = acct->proto->filter_text(msg);
+				ret = xstrncpy(buf, msg, len);
+				free(msg);
+			}
+			break;
+
+		case 'H':
+			if (acct->userhost != NULL) {
+				char *host = acct->proto->filter_text(acct->userhost);
+				ret = xstrncpy(buf, host, len);
+				free(host);
+			}
+			break;
+
+		default:
+			return (-1);
+	}
+
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
+}
+
+static int format_chat_recv(char opt, char *buf, size_t len, va_list ap) {
+	struct pork_acct *acct = va_arg(ap, struct pork_acct *);
+	struct chatroom *chat = va_arg(ap, struct chatroom *);
+	char *dest = va_arg(ap, char *);
+	char *src = va_arg(ap, char *);
+	char *src_uhost = va_arg(ap, char *);
+	char *msg = va_arg(ap, char *);
+
 	int ret = 0;
 
 	switch (opt) {
 		/* Message source */
 		case 'N':
+			if (src != NULL)
+				ret = xstrncpy(buf, buddy_name(acct, src), len);
+			break;
+
 		case 'n':
-			if (xstrncpy(buf, source, len) == -1)
-				ret = -1;
+			if (src != NULL)
+				ret = xstrncpy(buf, src, len);
 			break;
 
 		/* Message destination */
 		case 'C':
+			if (chat != NULL)
+				ret = xstrncpy(buf, chat->title_quoted, len);
+			break;
+
 		case 'c':
-			if (xstrncpy(buf, dest, len) == -1)
-				ret = -1;
+			if (dest != NULL) {
+				dest = acct->proto->filter_text(dest);
+				ret = xstrncpy(buf, dest, len);
+				free(dest);
+			}
 			break;
 
 		/* Message text */
-		case 'm':
 		case 'M':
-			if (xstrncpy(buf, msg, len) == -1)
-				ret = -1;
+		case 'm':
+			if (msg != NULL) {
+				msg = acct->proto->filter_text(msg);
+				ret = xstrncpy(buf, msg, len);
+				free(msg);
+			}
+			break;
+
+		case 'H':
+			if (acct->userhost != NULL) {
+				char *host = acct->proto->filter_text(acct->userhost);
+				ret = xstrncpy(buf, host, len);
+				free(host);
+			}
+			break;
+
+		case 'h':
+			if (src_uhost != NULL) {
+				src_uhost = acct->proto->filter_text(src_uhost);
+				ret = xstrncpy(buf, src_uhost, len);
+				free(src_uhost);
+			}
 			break;
 
 		default:
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_chat_info(char opt, char *buf, size_t len, va_list ap) {
-	char *name = va_arg(ap, char *);
-	char *action = va_arg(ap, char *);
-	char *room = va_arg(ap, char *);
-	int ret = 0;
-
-	switch (opt) {
-		case 'n':
-		case 'N':
-			if (xstrncpy(buf, name, len) == -1)
-				ret = -1;
-			break;
-
-		/* Action (join, part, etc) */
-		case 'a':
-		case 'A':
-			if (xstrncpy(buf, action, len) == -1)
-				ret = -1;
-			break;
-
-		case 'r':
-		case 'R':
-			if (xstrncpy(buf, room, len) == -1)
-				ret = -1;
-			break;
-
-		default:
-			return (-1);
-	}
-
-	return (ret);
-}
-
-static int format_chat_arginfo(char opt, char *buf, size_t len, va_list ap) {
-	char *name = va_arg(ap, char *);
-	char *room = va_arg(ap, char *);
-	char *altname = va_arg(ap, char *);
+	struct pork_acct *acct = va_arg(ap, struct pork_acct *);
+	struct chatroom *chat = va_arg(ap, struct chatroom *);
+	char *chat_nameq = va_arg(ap, char *);
+	char *src = va_arg(ap, char *);
+	char *dst = va_arg(ap, char *);
 	char *msg = va_arg(ap, char *);
 	int ret = 0;
 
 	switch (opt) {
+		/* Screen name of user who the action originated from */
 		case 'n':
 		case 'N':
-			if (xstrncpy(buf, name, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, buddy_name(acct, src), len);
 			break;
 
-		case 'u':
+		/* Destination, if applicable */
+		case 'D':
+		case 'd':
+			if (dst != NULL) {
+				dst = acct->proto->filter_text(dst);
+				ret = xstrncpy(buf, dst, len);
+				free(dst);
+			}
+			break;
+
+		/* Chat name (quoted) */
+		case 'r':
+		case 'R':
+			ret = xstrncpy(buf, chat_nameq, len);
+			break;
+
+		/* Chat name (full, quoted) */
 		case 'U':
-			if (xstrncpy(buf, altname, len) == -1)
-				ret = -1;
+		case 'u':
+			if (chat != NULL)
+				ret = xstrncpy(buf, chat->title_full_quoted, len);
+
+		/* Source's userhost (if available) */
+		case 'H':
+			if (acct != NULL && chat != NULL && src != NULL) {
+				struct chat_user *chat_user;
+
+				chat_user = chat_find_user(acct, chat, src);
+				if (chat_user != NULL && chat_user->host != NULL) {
+					char *host = acct->proto->filter_text(chat_user->host);
+					ret = xstrncpy(buf, host, len);
+					free(host);
+				}
+			}
+			break;
+
+		/* Dest's userhost (if available) */
+		case 'h':
+			if (acct != NULL && chat != NULL && dst != NULL) {
+				struct chat_user *chat_user;
+
+				chat_user = chat_find_user(acct, chat, dst);
+				if (chat_user != NULL && chat_user->host != NULL) {
+					char *host = acct->proto->filter_text(chat_user->host);
+					ret = xstrncpy(buf, host, len);
+					free(host);
+				}
+			}
 			break;
 
 		case 'm':
 		case 'M':
-		{
-			char *chat_msg = msg;
-			if (chat_msg == NULL)
-				chat_msg = " ";
+			if (msg == NULL)
+				break;
 
-			if (xstrncpy(buf, chat_msg, len) == -1)
-				ret = -1;
-
-			break;
-		}
-
-		case 'r':
-		case 'R':
-			if (xstrncpy(buf, room, len) == -1)
-				ret = -1;
+			msg = acct->proto->filter_text(msg);
+			ret = xstrncpy(buf, msg, len);
+			free(msg);
 			break;
 
 		default:
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_warning(char opt, char *buf, size_t len, va_list ap) {
@@ -520,32 +803,27 @@ static int format_warning(char opt, char *buf, size_t len, va_list ap) {
 	switch (opt) {
 		case 'n':
 		case 'N':
-			if (xstrncpy(buf, you, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, you, len);
 			break;
 
 		case 'u':
 		case 'U':
-			if (xstrncpy(buf, warner, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, warner, len);
 			break;
 
 		case 'w':
 		case 'W':
-		{
-			int r_code;
-
-			r_code = snprintf(buf, len, "%u", warn_level);
-			if (r_code < 0 || (size_t) r_code >= len)
-				ret = -1;
+			ret = snprintf(buf, len, "%u", warn_level);
 			break;
-		}
 
 		default:
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_blist_idle(char opt, char *buf, size_t len, va_list ap) {
@@ -554,14 +832,17 @@ static int format_blist_idle(char opt, char *buf, size_t len, va_list ap) {
 
 	switch (opt) {
 		case 'I':
-			time_to_str(buddy->idle_time, buf, len);
+			ret = time_to_str(buddy->idle_time, buf, len);
 			break;
 
 		default:
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_blist_warn(char opt, char *buf, size_t len, va_list ap) {
@@ -570,20 +851,17 @@ static int format_blist_warn(char opt, char *buf, size_t len, va_list ap) {
 
 	switch (opt) {
 		case 'W':
-		{
-			int r_code;
-
-			r_code = snprintf(buf, len, "%u", buddy->warn_level);
-			if (r_code < 0 || (size_t) r_code >= len)
-				ret = -1;
+			ret = snprintf(buf, len, "%u", buddy->warn_level);
 			break;
-		}
 
 		default:
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_blist_buddy_label(char opt, char *buf, size_t len, va_list ap)
@@ -593,8 +871,7 @@ static int format_blist_buddy_label(char opt, char *buf, size_t len, va_list ap)
 
 	switch (opt) {
 		case 'B':
-		case 'b':
-		{
+		case 'b': {
 			char *status_text;
 
 			if (buddy->status == STATUS_ACTIVE)
@@ -606,16 +883,13 @@ static int format_blist_buddy_label(char opt, char *buf, size_t len, va_list ap)
 			else
 				status_text = "%p?%x";
 
-			if (xstrncpy(buf, status_text, len) == -1)
-				ret = -1;
-
+			ret = xstrncpy(buf, status_text, len);
 			break;
 		}
 
 		case 'N':
 		case 'n':
-			if (xstrncpy(buf, buddy->name, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, buddy->name, len);
 			break;
 
 		case 'I':
@@ -640,7 +914,10 @@ static int format_blist_buddy_label(char opt, char *buf, size_t len, va_list ap)
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) 	ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_blist_group_label(char opt, char *buf, size_t len, va_list ap)
@@ -650,8 +927,7 @@ static int format_blist_group_label(char opt, char *buf, size_t len, va_list ap)
 	int ret = 0;
 
 	switch (opt) {
-		case 'E':
-		{
+		case 'E': {
 			char *expand_str;
 
 			if (node != NULL) {
@@ -669,52 +945,46 @@ static int format_blist_group_label(char opt, char *buf, size_t len, va_list ap)
 					opt_get_str(OPT_TEXT_BLIST_GROUP_COLLAPSED);
 			}
 
-			if (xstrncpy(buf, expand_str, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, expand_str, len);
 			break;
 		}
 
 		case 'N':
 		case 'n':
-			if (xstrncpy(buf, group->name, len) == -1)
-				ret = -1;
+			ret = xstrncpy(buf, group->name, len);
 			break;
 
 		case 'T':
 		case 't':
-		{
-			int r_code;
-
-			r_code = snprintf(buf, len, "%u", group->num_members);
-			if (r_code < 0 || (size_t) r_code >= len)
-				ret = -1;
+			ret = snprintf(buf, len, "%u", group->num_members);
 			break;
-		}
 
 		case 'O':
 		case 'o':
-		{
-			int r_code;
-
-			r_code = snprintf(buf, len, "%u", group->num_online);
-			if (r_code < 0 || (size_t) r_code >= len)
-				ret = -1;
+			ret = snprintf(buf, len, "%u", group->num_online);
 			break;
-		}
 
 		default:
 			return (-1);
 	}
 
-	return (ret);
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
 }
 
 static int format_file_transfer(char opt, char *buf, size_t len, va_list ap) {
 	struct file_transfer *xfer = va_arg(ap, struct file_transfer *);
 	struct pork_acct *acct = xfer->acct;
-	int ret;
+	int ret = 0;
 
 	switch (opt) {
+		/* Starting offset */
+		case 'O':
+			ret = snprintf(buf, len, "%lld", xfer->start_offset);
+			break;
+
 		/* File transfer reference number */
 		case 'I':
 			ret = snprintf(buf, len, "%u", xfer->refnum);
@@ -767,7 +1037,7 @@ static int format_file_transfer(char opt, char *buf, size_t len, va_list ap) {
 
 		/* File size */
 		case 'S':
-			ret = snprintf(buf, len, "%lld", xfer->file_len);
+			ret = snprintf(buf, len, "%lld", xfer->file_len - xfer->start_offset);
 			break;
 
 		/* Number of bytes sent */
@@ -800,42 +1070,117 @@ static int format_file_transfer(char opt, char *buf, size_t len, va_list ap) {
 	return (0);
 }
 
+static int format_whois(char opt, char *buf, size_t len, va_list ap) {
+	char *user = va_arg(ap, char*);
+	u_int32_t warn_level = va_arg(ap, unsigned long int);
+	u_int32_t idle_time = va_arg(ap, unsigned long int);
+	u_int32_t online_since = va_arg(ap, unsigned long int);
+	u_int32_t member_since = va_arg(ap, unsigned long int);
+	int ret = 0;
+
+	switch (opt) {
+		/* Name */
+		case 'N':
+		case 'n':
+			ret = xstrncpy(buf, user, len);
+			break;
+
+		/* Online since */
+		case 'S':
+		case 's':
+			ret = date_to_str(online_since, buf, len);
+			break;
+
+		/* Idle time */
+		case 'I':
+		case 'i':
+			ret = time_to_str(idle_time, buf, len);
+			break;
+
+		/* Member since */
+		case 'D':
+		case 'd':
+			ret = date_to_str(member_since, buf, len);
+			break;
+
+		/* Warn level */
+		case 'W':
+		case 'w':
+			ret = snprintf(buf, len, "%u", warn_level);
+			break;
+
+		default:
+			return (-1);
+	}
+
+	if (ret < 0 || (size_t) ret >= len)
+		return (-1);
+
+	return (0);
+}
+
 static int (*const format_handler[])(char, char *, size_t, va_list) = {
+	format_msg_recv,			/* OPT_FORMAT_ACTION_RECV			*/
+	format_msg_recv,			/* OPT_FORMAT_ACTION_RECV_STATUS	*/
+	format_msg_send,			/* OPT_FORMAT_ACTION_SEND			*/
+	format_msg_send,			/* OPT_FORMAT_ACTION_SEND_STATUS	*/
 	format_blist_buddy_label,	/* OPT_FORMAT_BLIST					*/
 	format_blist_group_label,	/* OPT_FORMAT_BLIST_GROUP			*/
 	format_blist_idle,			/* OPT_FORMAT_BLIST_IDLE			*/
 	format_blist_warn,			/* OPT_FORMAT_BLIST_WARN			*/
-	format_chat_arginfo,		/* OPT_FORMAT_CHAT_CREATE			*/
-	format_chat_arginfo,		/* OPT_FORMAT_CHAT_IGNORE			*/
-	format_chat_info,			/* OPT_FORMAT_CHAT_INFO				*/
-	format_chat_arginfo,		/* OPT_FORMAT_CHAT_INVITE			*/
-	format_chat_arginfo,		/* OPT_FORMAT_CHAT_KICK				*/
-	format_chat_msg,			/* OPT_FORMAT_CHAT_RECV				*/
-	format_chat_msg,			/* OPT_FORMAT_CHAT_RECV_NOTICE		*/
-	format_chat_msg,			/* OPT_FORMAT_CHAT_SEND				*/
-	format_chat_msg,			/* OPT_FORMAT_CHAT_SEND_NOTICE		*/
-	format_chat_msg,			/* OPT_FORMAT_CHAT_TOPIC			*/
-	format_chat_arginfo,		/* OPT_FORMAT_CHAT_UNIGNORE			*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_CREATE			*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_IGNORE			*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_INVITE			*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_JOIN				*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_KICK				*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_LEAVE			*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_MODE				*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_QUIT				*/
+	format_chat_recv,			/* OPT_FORMAT_CHAT_RECV				*/
+	format_chat_recv,			/* OPT_FORMAT_CHAT_RECV_ACTION		*/
+	format_chat_recv,			/* OPT_FORMAT_CHAT_RECV_NOTICE		*/
+	format_chat_send,			/* OPT_FORMAT_CHAT_SEND				*/
+	format_chat_send,			/* OPT_FORMAT_CHAT_SEND_ACTION		*/
+	format_chat_send,			/* OPT_FORMAT_CHAT_SEND_NOTICE		*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_TOPIC			*/
+	format_chat_info,			/* OPT_FORMAT_CHAT_UNIGNORE			*/
+	format_file_transfer,		/* OPT_FORMAT_FILE_CANCEL_LOCAL		*/
+	format_file_transfer,		/* OPT_FORMAT_FILE_CANCEL_REMOTE	*/
+	format_file_transfer,		/* OPT_FORMAT_FILE_LOST				*/
 	format_file_transfer,		/* OPT_FORMAT_FILE_RECV_ACCEPT		*/
 	format_file_transfer,		/* OPT_FORMAT_FILE_RECV_ASK			*/
 	format_file_transfer,		/* OPT_FORMAT_FILE_RECV_COMPLETE	*/
+	format_file_transfer,		/* OPT_FORMAT_FILE_RECV_RESUME		*/
 	format_file_transfer,		/* OPT_FORMAT_FILE_SEND_ACCEPT		*/
 	format_file_transfer,		/* OPT_FORMAT_FILE_SEND_ASK			*/
 	format_file_transfer,		/* OPT_FORMAT_FILE_SEND_COMPLETE	*/
-	format_privmsg,				/* OPT_FORMAT_IM_RECV				*/
-	format_privmsg,				/* OPT_FORMAT_IM_RECV_AUTO			*/
-	format_privmsg,				/* OPT_FORMAT_IM_SEND				*/
-	format_privmsg,				/* OPT_FORMAT_IM_SEND_AUTO			*/
-	format_privmsg,				/* OPT_FORMAT_NOTICE_RECV			*/
-	format_privmsg,				/* OPT_FORMAT_NOTICE_SEND			*/
+	format_file_transfer,		/* OPT_FORMAT_FILE_SEND_RESUME		*/
+	format_msg_recv,			/* OPT_FORMAT_IM_RECV				*/
+	format_msg_recv,			/* OPT_FORMAT_IM_RECV_AUTO			*/
+	format_msg_recv,			/* OPT_FORMAT_IM_RECV_STATUS		*/
+	format_msg_send,			/* OPT_FORMAT_IM_SEND				*/
+	format_msg_send,			/* OPT_FORMAT_IM_SEND_AUTO			*/
+	format_msg_send,			/* OPT_FORMAT_IM_SEND_STATUS		*/
+	format_msg_recv,			/* OPT_FORMAT_NOTICE_RECV			*/
+	format_msg_recv,			/* OPT_FORMAT_NOTICE_RECV_STATUS	*/
+	format_msg_send,			/* OPT_FORMAT_NOTICE_SEND			*/
+	format_msg_send,			/* OPT_FORMAT_NOTICE_SEND_STATUS	*/
 	format_status,				/* OPT_FORMAT_STATUS				*/
 	format_status_activity,		/* OPT_FORMAT_STATUS_ACTIVITY		*/
+	format_status,				/* OPT_FORMAT_STATUS_CHAT			*/
 	format_status_held,			/* OPT_FORMAT_STATUS_HELD			*/
 	format_status_idle,			/* OPT_FORMAT_STATUS_IDLE			*/
 	format_status_timestamp,	/* OPT_FORMAT_STATUS_TIMESTAMP		*/
 	format_status_typing,		/* OPT_FORMAT_STATUS_TYPING			*/
 	format_status_warn,			/* OPT_FORMAT_STATUS_WARN			*/
 	format_warning,				/* OPT_FORMAT_WARN					*/
+	format_whois,				/* OPT_FORMAT_WHOIS_AWAY			*/
+	format_whois,				/* OPT_FORMAT_WHOIS_IDLE			*/
+	format_whois,				/* OPT_FORMAT_WHOIS_MEMBER			*/
+	format_whois,				/* OPT_FORMAT_WHOIS_NAME			*/
+	format_whois,				/* OPT_FORMAT_WHOIS_SIGNON			*/
+	format_whois,				/* OPT_FORMAT_WHOIS_USERINFO		*/
+	format_whois,				/* OPT_FORMAT_WHOIS_WARNLEVEL		*/
 };
 
 /*
@@ -850,13 +1195,13 @@ int fill_format_str(int type, char *buf, size_t len, ...) {
 	int (*handler)(char, char *, size_t, va_list);
 
 	format = opt_get_str(type);
-	if (format == NULL)
+	if (format == NULL) {
+		debug("unknown format str: %d", type);
 		return (-1);
+	}
 
-	handler = format_handler[type - OPT_FORMAT_OFFSET];
-
-	va_start(ap, len);
 	--len;
+	handler = format_handler[type - OPT_FORMAT_OFFSET];
 
 	while (*format != '\0' && i < len) {
 		if (*format == FORMAT_VARIABLE) {
@@ -866,7 +1211,11 @@ int fill_format_str(int type, char *buf, size_t len, ...) {
 			format++;
 
 			result[0] = '\0';
+
+			va_start(ap, len);
 			ret_code = handler(*format, result, sizeof(result), ap);
+			va_end(ap);
+
 			if (ret_code == 0) {
 				/*
 				** The subsitution was successful.
@@ -887,8 +1236,8 @@ int fill_format_str(int type, char *buf, size_t len, ...) {
 				break;
 			} else {
 				/*
-				** They specified an invalid option -- treat
-				** their '$' as a literal.
+				** Unknown variable -- treat
+				** their FORMAT_VARIABLE as a literal.
 				*/
 
 				if (i > len - 2)
@@ -905,7 +1254,6 @@ int fill_format_str(int type, char *buf, size_t len, ...) {
 
 	buf[i] = '\0';
 
-	va_end(ap);
 	return (0);
 }
 
@@ -924,6 +1272,7 @@ void format_apply_justification(char *buf, chtype *ch, size_t len) {
 			right = &p[2];
 			break;
 		}
+		p++;
 	}
 
 	len_left = plaintext_to_cstr(buf, ch, len);
