@@ -1,6 +1,6 @@
 /*
 ** pork_misc.h - miscellaneous functions.
-** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2002-2004 Ryan McCabe <ryan@numb.org>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -17,8 +17,8 @@ void set_default_win_opts(WINDOW *win);
 
 void resize_terminal(int rows, int cols);
 int normalize(char *dest, const char *src, size_t len);
-void time_to_str(u_int32_t timespec, char *buf, size_t len);
-void time_to_str_full(u_int32_t timespec, char *buf, size_t len);
+int time_to_str(u_int32_t timespec, char *buf, size_t len);
+int time_to_str_full(u_int32_t timespec, char *buf, size_t len);
 int date_to_str(time_t timespec, char *buf, size_t len);
 int wgetinput(WINDOW *win);
 
