@@ -1,6 +1,6 @@
 /*
 ** pork_perl_xs.c - Perl scripting support
-** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2002-2004 Ryan McCabe <ryan@numb.org>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -21,7 +21,10 @@
 
 #undef PACKAGE
 #undef instr
-#define HAS_BOOL
+
+#ifndef HAS_BOOL
+#	define HAS_BOOL
+#endif
 
 #include <EXTERN.h>
 #include <perl.h>
@@ -59,6 +62,30 @@
 
 extern struct screen screen;
 
+XS(PORK_run_cmd) {
+	char *command_str;
+	size_t notused;
+	u_int32_t i = 0;
+	dXSARGS;
+
+	(void) cv;
+
+	if (items < 1)
+		XSRETURN_IV(-1);
+
+	while (items-- > 0) {
+		command_str = SvPV(ST(i), notused);
+		i++;
+
+		if (command_str == NULL)
+			continue;
+
+		run_command(command_str);
+	}
+
+	XSRETURN_IV(0);
+}
+
 XS(PORK_connect) {
 	char *username = NULL;
 	size_t notused;
@@ -73,6 +100,25 @@ XS(PORK_connect) {
 	XSRETURN_IV(0);
 }
 
+XS(PORK_input_send) {
+	char *args;
+	size_t notused;
+	dXSARGS;
+
+	(void) cv;
+
+	if (items == 0)
+		XSRETURN_IV(-1);
+
+	args = SvPV(ST(0), notused);
+
+	if (args == NULL)
+		XSRETURN_IV(-1);
+
+	cmd_input_send(args);
+	XSRETURN_IV(0);
+}
+
 XS(PORK_disconnect) {
 	struct pork_acct *acct;
 	char *reason = NULL;
@@ -101,6 +147,14 @@ XS(PORK_disconnect) {
 	XSRETURN_IV(pork_acct_del_refnum(acct->refnum, reason));
 }
 
+XS(PORK_cur_user) {
+	dXSARGS;
+
+	(void) cv;
+	(void) items;
+	XSRETURN_PV(cur_window()->owner->username);
+}
+
 XS(PORK_err_msg) {
 	size_t notused;
 	char *msg;
@@ -119,6 +173,29 @@ XS(PORK_err_msg) {
 	XSRETURN_IV(0);
 }
 
+XS(PORK_prompt_user) {
+	dXSARGS;
+	char buf[4096];
+	int ret;
+	char *prompt = NULL;
+
+	(void) cv;
+
+	if (items > 0) {
+		size_t notused;
+
+		prompt = SvPV(ST(0), notused);
+	}
+
+	screen_prompt_user(prompt, buf, sizeof(buf));
+	ret = strlen(buf);
+
+	if (ret <= 0)
+		XSRETURN_UNDEF;
+
+	XSRETURN_PV(buf);
+}
+
 XS(PORK_status_msg) {
 	size_t notused;
 	char *msg;
@@ -226,6 +303,33 @@ XS(PORK_quit) {
 	XSRETURN_IV(0);
 }
 
+XS(PORK_quote) {
+	struct pork_acct *acct;
+	char *str;
+	size_t notused;
+	dXSARGS;
+
+	(void) cv;
+
+	str = SvPV(ST(0), notused);
+	if (str == NULL)
+		XSRETURN_IV(-1);
+
+	if (items > 1) {
+		u_int32_t acct_refnum = SvIV(ST(1));
+
+		acct = pork_acct_get_data(acct_refnum);
+		if (acct == NULL)
+			XSRETURN_IV(-1);
+	} else
+		acct = cur_window()->owner;
+
+	if (acct->proto->quote == NULL)
+		XSRETURN_IV(-1);
+
+	XSRETURN_IV(acct->proto->quote(acct, str));
+}
+
 XS(PORK_refresh) {
 	dXSARGS;
 
@@ -436,10 +540,11 @@ XS(PORK_alias_get) {
 	if (alias == NULL)
 		XSRETURN_EMPTY;
 
-	XST_mPV(0, alias->orig);
+	SP -= items;
+	XPUSHs(sv_2mortal(newSVpv(alias->orig, 0)));
 
 	if (alias->args != NULL) {
-		XST_mPV(1, alias->args);
+		XPUSHs(sv_2mortal(newSVpv(alias->args, 0)));
 		XSRETURN(2);
 	}
 
@@ -559,11 +664,12 @@ XS(PORK_get_acct_list) {
 	if (items != 0 || cur == NULL)
 		XSRETURN_EMPTY;
 
+	SP -= items;
 	do {
 		struct pork_acct *acct = cur->data;
 
 		if (acct->can_connect) {
-			XST_mPV(args, acct->username);
+			XPUSHs(sv_2mortal(newSVpv(acct->username, 0)));
 			args++;
 		}
 		cur = cur->next;
@@ -850,7 +956,7 @@ XS(PORK_set_away) {
 	if (items > 2)
 		XSRETURN_IV(-1);
 
-	if (items >= 1)
+	if (items >= 1 && SvTRUE(ST(0)))
 		msg = SvPV(ST(0), notused);
 
 	if (items == 2) {
@@ -1186,13 +1292,12 @@ XS(PORK_buddy_get_groups) {
 	if (acct->buddy_pref == NULL)
 		XSRETURN_EMPTY;
 
-	cur = acct->buddy_pref->group_list;
-	while (cur != NULL) {
+	SP -= items;
+	for (cur = acct->buddy_pref->group_list ; cur != NULL ; cur = cur->next) {
 		struct bgroup *group = cur->data;
 
-		XST_mPV(args, group->name);
+		XPUSHs(sv_2mortal(newSVpv(group->name, 0)));
 		args++;
-		cur = cur->next;
 	}
 
 	XSRETURN(args);
@@ -1218,11 +1323,10 @@ XS(PORK_buddy_get_block) {
 	if (acct->buddy_pref == NULL)
 		XSRETURN_EMPTY;
 
-	cur = acct->buddy_pref->block_list;
-	while (cur != NULL) {
-		XST_mPV(args, cur->data);
+	SP -= items;
+	for (cur = acct->buddy_pref->block_list ; cur != NULL ; cur = cur->next) {
+		XPUSHs(sv_2mortal(newSVpv(cur->data, 0)));
 		args++;
-		cur = cur->next;
 	}
 
 	XSRETURN(args);
@@ -1248,11 +1352,10 @@ XS(PORK_buddy_get_permit) {
 	if (acct->buddy_pref == NULL)
 		XSRETURN_EMPTY;
 
-	cur = acct->buddy_pref->permit_list;
-	while (cur != NULL) {
-		XST_mPV(args, cur->data);
+	SP -= items;
+	for (cur = acct->buddy_pref->permit_list ; cur != NULL ; cur = cur->next) {
+		XPUSHs(sv_2mortal(newSVpv(cur->data, 0)));
 		args++;
-		cur = cur->next;
 	}
 
 	XSRETURN(args);
@@ -1289,13 +1392,12 @@ XS(PORK_buddy_get_group_members) {
 	if (bgroup == NULL)
 		XSRETURN_EMPTY;
 
-	cur = bgroup->member_list;
-	while (cur != NULL) {
+	SP -= items;
+	for (cur = bgroup->member_list ; cur != NULL ; cur = cur->next) {
 		struct buddy *buddy = cur->data;
 
-		XST_mPV(args, buddy->nname);
+		XPUSHs(sv_2mortal(newSVpv(buddy->nname, 0)));
 		args++;
-		cur = cur->next;
 	}
 
 	XSRETURN(args);
@@ -1609,6 +1711,35 @@ XS(PORK_win_find_name) {
 	XSRETURN_IV(imwindow->refnum);
 }
 
+XS(PORK_win_target) {
+	u_int32_t win_refnum;
+	struct imwindow *imwindow;
+	struct pork_acct *acct;
+	dXSARGS;
+
+	(void) cv;
+
+	if (items == 0)
+		XSRETURN_PV(cur_window()->target);
+
+	win_refnum = SvIV(ST(0));
+
+	if (items > 1) {
+		u_int32_t acct_refnum = SvIV(ST(1));
+
+		acct = pork_acct_get_data(acct_refnum);
+		if (acct == NULL)
+			XSRETURN_UNDEF;
+	} else
+		acct = cur_window()->owner;
+
+	imwindow = imwindow_find_refnum(win_refnum);
+	if (imwindow == NULL)
+		XSRETURN_UNDEF;
+
+	XSRETURN_PV(imwindow->target);
+}
+
 XS(PORK_win_find_target) {
 	char *str;
 	size_t notused;
@@ -2022,17 +2153,19 @@ XS(PORK_blist_cursor) {
 	if (cell == NULL)
 		XSRETURN_EMPTY;
 
+	SP -= items;
+
 	if (cell->type == TYPE_FLAT_CELL) {
 		struct buddy *buddy = cell->data;
 
-		XST_mPV(0, buddy->nname);
+		XPUSHs(sv_2mortal(newSVpv(buddy->nname, 0)));
 	} else {
 		struct bgroup *group = cell->data;
 
-		XST_mPV(0, group->name);
+		XPUSHs(sv_2mortal(newSVpv(group->name, 0)));
 	}
 
-	XST_mIV(1, cell->type);
+	XPUSHs(sv_2mortal(newSViv(cell->type)));
 	XSRETURN(2);
 }
 
@@ -2431,6 +2564,125 @@ XS(PORK_event_purge) {
 ** /chat commands.
 */
 
+XS(PORK_chat_ban) {
+	char *chat_name;
+	char *ban_user;
+	size_t notused;
+	struct pork_acct *acct;
+	struct chatroom *chat;
+	dXSARGS;
+
+	(void) cv;
+
+	if (items != 2 && items != 3)
+		XSRETURN_IV(-1);
+
+	chat_name = SvPV(ST(0), notused);
+	ban_user = SvPV(ST(1), notused);
+
+	if (chat_name == NULL || ban_user == NULL)
+		XSRETURN_IV(-1);
+
+	if (items == 3) {
+		u_int32_t acct_refnum = SvIV(ST(2));
+
+		acct = pork_acct_get_data(acct_refnum);
+		if (acct == NULL)
+			XSRETURN_IV(-1);
+	} else
+		acct = cur_window()->owner;
+
+	chat = chat_find(acct, chat_name);
+	if (chat == NULL)
+		XSRETURN_IV(-1);
+
+	XSRETURN_IV(chat_ban(acct, chat, ban_user));
+}
+
+XS(PORK_chat_kick) {
+	char *chat_name;
+	char *user;
+	char *reason = NULL;
+	size_t notused;
+	struct pork_acct *acct;
+	struct chatroom *chat;
+	dXSARGS;
+
+	(void) cv;
+
+	if (items < 2)
+		XSRETURN_IV(-1);
+
+	chat_name = SvPV(ST(0), notused);
+	user = SvPV(ST(1), notused);
+
+	if (items > 2)
+		reason = SvPV(ST(2), notused);
+
+	if (chat_name == NULL || user == NULL)
+		XSRETURN_IV(-1);
+
+	if (reason == NULL)
+		reason = "";
+
+	if (items == 4) {
+		u_int32_t acct_refnum = SvIV(ST(3));
+
+		acct = pork_acct_get_data(acct_refnum);
+		if (acct == NULL)
+			XSRETURN_IV(-1);
+	} else
+		acct = cur_window()->owner;
+
+	chat = chat_find(acct, chat_name);
+	if (chat == NULL)
+		XSRETURN_IV(-1);
+
+	XSRETURN_IV(chat_kick(acct, chat, user, reason));
+}
+
+XS(PORK_chat_topic) {
+	char *chat_name;
+	char *topic;
+	size_t notused;
+	struct pork_acct *acct;
+	struct chatroom *chat;
+	dXSARGS;
+
+	(void) cv;
+
+	if (items != 2 && items != 3)
+		XSRETURN_IV(-1);
+
+	chat_name = SvPV(ST(0), notused);
+	topic = SvPV(ST(1), notused);
+
+	if (chat_name == NULL)
+		XSRETURN_IV(-1);
+
+	if (items == 3) {
+		u_int32_t acct_refnum = SvIV(ST(2));
+
+		acct = pork_acct_get_data(acct_refnum);
+		if (acct == NULL)
+			XSRETURN_IV(-1);
+	} else
+		acct = cur_window()->owner;
+
+	chat = chat_find(acct, chat_name);
+	if (chat == NULL)
+		XSRETURN_IV(-1);
+
+	if (topic == NULL || *topic == '\0') {
+		if (chat->topic != NULL)
+			XSRETURN_PV(chat->topic);
+		else
+			XSRETURN_UNDEF;
+	}
+
+	XSRETURN_IV(chat_set_topic(acct, chat, topic));
+}
+
 XS(PORK_chat_get_list) {
 	dXSARGS;
 	dlist_t *cur;
@@ -2448,14 +2700,11 @@ XS(PORK_chat_get_list) {
 	} else
 		acct = cur_window()->owner;
 
+	SP -= items;
 	for (cur = acct->chat_list ; cur != NULL ; cur = cur->next) {
-		struct chatroom *chat;
-		char buf[256];
+		struct chatroom *chat = cur->data;
 
-		chat = cur->data;
-		snprintf(buf, sizeof(buf), "%s", chat->title);
-
-		XST_mPV(i, buf);
+		XPUSHs(sv_2mortal(newSVpv(chat->title, 0)));
 		i++;
 	}
 
@@ -2500,20 +2749,25 @@ XS(PORK_chat_get_users) {
 	dlist_t *node;
 	struct pork_acct *acct;
 	struct chatroom *chat;
+	u_int32_t flags = 0;
 	u_int32_t args = 0;
+	u_int32_t invert = 0;
 	dXSARGS;
 
 	(void) cv;
 
-	if (items != 1 && items != 2)
+	if (items < 3)
 		XSRETURN_EMPTY;
 
 	chat_name = SvPV(ST(0), notused);
 	if (chat_name == NULL)
 		XSRETURN_EMPTY;
 
-	if (items == 2) {
-		u_int32_t acct_refnum = SvIV(ST(1));
+	flags = SvIV(ST(1));
+	invert = SvIV(ST(2));
+
+	if (items == 4) {
+		u_int32_t acct_refnum = SvIV(ST(3));
 
 		acct = pork_acct_get_data(acct_refnum);
 		if (acct == NULL)
@@ -2525,14 +2779,14 @@ XS(PORK_chat_get_users) {
 	if (chat == NULL)
 		XSRETURN_EMPTY;
 
-	node = chat->user_list;
-	while (node != NULL) {
-		struct buddy *buddy = node->data;
+	SP -= items;
+	for (node = chat->user_list ; node != NULL ; node = node->next) {
+		struct chat_user *user = node->data;
 
-		XST_mPV(args, buddy->nname);
-		args++;
-
-		node = node->next;
+		if (!flags || (!!(user->status & flags)) ^ invert) {
+			XPUSHs(sv_2mortal(newSVpv(user->nname, 0)));
+			args++;
+		}
 	}
 
 	XSRETURN(args);
@@ -2689,7 +2943,7 @@ XS(PORK_chat_send) {
 	if (chat == NULL)
 		XSRETURN_IV(-1);
 
-	XSRETURN_IV(chat_send_msg(acct, chat, msg));
+	XSRETURN_IV(chat_send_msg(acct, chat, chat_name, msg));
 }
 
 XS(PORK_chat_unignore) {
@@ -2721,3 +2975,40 @@ XS(PORK_chat_unignore) {
 
 	XSRETURN_IV(chat_unignore(acct, name, dest));
 }
+
+XS(PORK_chat_target) {
+	char *name = NULL;
+	size_t notused;
+	struct pork_acct *acct;
+	struct chatroom *chat;
+	dXSARGS;
+
+	(void) cv;
+
+	if (items > 0) {
+		name = SvPV(ST(0), notused);
+		if (name == NULL)
+			XSRETURN_UNDEF;
+	} else {
+		if (cur_window()->type != WIN_TYPE_CHAT)
+			XSRETURN_UNDEF;
+
+		chat = cur_window()->data;
+		XSRETURN_PV(chat->title);
+	}
+
+	if (items == 2) {
+		u_int32_t acct_refnum = SvIV(ST(1));
+
+		acct = pork_acct_get_data(acct_refnum);
+		if (acct == NULL)
+			XSRETURN_UNDEF;
+	} else
+		acct = cur_window()->owner;
+
+	chat = chat_find(acct, name);
+	if (chat == NULL)
+		XSRETURN_UNDEF;
+
+	XSRETURN_PV(chat->title);
+}
