@@ -1,6 +1,6 @@
 /*
 ** pork_proto.c
-** Copyright (C) 2003 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2003-2004 Ryan McCabe <ryan@numb.org>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -12,6 +12,7 @@
 #include <unistd.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <sys/types.h>
 
 #include <pork.h>
@@ -29,7 +30,10 @@ inline struct pork_proto *proto_get(int protocol) {
 	return (proto_table[protocol + 1]);
 }
 
-int proto_new(int protocol, int (*init_func)(struct pork_proto *)) {
+int proto_new(	int protocol,
+				const char *name,
+				int (*init_func)(struct pork_proto *))
+{
 	struct pork_proto *new_proto;
 	int ret;
 
@@ -41,6 +45,7 @@ int proto_new(int protocol, int (*init_func)(struct pork_proto *)) {
 
 	new_proto = xcalloc(1, sizeof(*new_proto));
 	new_proto->protocol = protocol;
+	xstrncpy(new_proto->name, name, sizeof(new_proto->name));
 
 	ret = init_func(new_proto);
 	if (ret != 0) {
@@ -52,10 +57,21 @@ int proto_new(int protocol, int (*init_func)(struct pork_proto *)) {
 	return (0);
 }
 
+int proto_get_num(const char *name) {
+	size_t i;
+
+	for (i = 0 ; i <= PROTO_MAX && proto_table[i] != NULL ; i++) {
+		if (!strcasecmp(proto_table[i]->name, name))
+			return (proto_table[i]->protocol);
+	}
+
+	return (-1);
+}
+
 void proto_destroy(void) {
 	size_t i;
 
-	for (i = 0 ; i < PROTO_MAX ; i++)
+	for (i = 0 ; i <= PROTO_MAX ; i++)
 		free(proto_table[i]);
 }
 
@@ -66,5 +82,5 @@ static int proto_init_null(struct pork_proto *proto) {
 }
 
 int proto_init(void) {
-	return (proto_new(PROTO_NULL, proto_init_null));
+	return (proto_new(PROTO_NULL, "NULL", proto_init_null));
 }
