@@ -1,6 +1,6 @@
 /*
 ** pork_queue.h - Generic queues
-** Copyright (C) 2003 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2003-2004 Ryan McCabe <ryan@numb.org>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -15,11 +15,12 @@ typedef struct {
 	dlist_t *tail;
 	u_int32_t entries;
 	u_int32_t max;
-} queue_t;
+} pork_queue_t;
 
-queue_t *queue_new(u_int32_t max_entries);
-int queue_add(queue_t *q, void *data);
-void *queue_get(queue_t *q);
-void queue_destroy(queue_t *q, void (*cleanup)(void *));
+pork_queue_t *queue_new(u_int32_t max_entries);
+int queue_add(pork_queue_t *q, void *data);
+int queue_putback_head(pork_queue_t *q, void *data);
+void *queue_get(pork_queue_t *q);
+void queue_destroy(pork_queue_t *q, void (*cleanup)(void *));
 
 #endif
