@@ -1,7 +1,7 @@
 /*
 ** pork_set.c - /SET command implementation.
-** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
-** Copyright (C) 2002-2003 Amber Adams <amber@ojnk.net>
+** Copyright (C) 2002-2004 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2002-2004 Amber Adams <amber@ojnk.net>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -14,6 +14,7 @@
 #include <ncurses.h>
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h>
 #include <time.h>
 #include <sys/time.h>
 
@@ -46,6 +47,8 @@ static int wopt_set_char(	struct imwindow *imwindow,
 
 static void opt_changed_prompt(void);
 
+static void scrollbuf_len_update(void);
+
 static void wopt_changed_histlen(struct imwindow *imwindow);
 static void wopt_changed_log(struct imwindow *imwindow);
 static void wopt_changed_logfile(struct imwindow *imwindow);
@@ -55,6 +58,7 @@ static void wopt_changed_scroll_on_output(struct imwindow *imwindow);
 static void wopt_changed_scroll_on_input(struct imwindow *imwindow);
 static void wopt_changed_show_blist(struct imwindow *imwindow);
 static void wopt_changed_timestamp(struct imwindow *imwindow);
+static void wopt_changed_wordwrap(struct imwindow *imwindow);
 
 /*
 ** Name:		The name of the option
@@ -74,6 +78,12 @@ struct global_var global_var[] = {
 		opt_set_bool,
 		NULL,
 		SET_BOOL(DEFAULT_AUTO_RECONNECT)
+	},{	"AUTO_REJOIN",
+		OPT_BOOL,
+		0,
+		opt_set_bool,
+		NULL,
+		SET_BOOL(DEFAULT_AUTO_REJOIN)
 	},{	"AUTOSEND_AWAY",
 		OPT_BOOL,
 		0,
@@ -86,6 +96,18 @@ struct global_var global_var[] = {
 		opt_set_str,
 		NULL,
 		SET_STR(DEFAULT_BANNER)
+	},{	"BEEP",
+		OPT_BOOL,
+		0,
+		opt_set_bool,
+		NULL,
+		SET_BOOL(DEFAULT_BEEP)
+	},{	"BEEP_MAX",
+		OPT_INT,
+		0,
+		opt_set_int,
+		NULL,
+		SET_INT(DEFAULT_BEEP_MAX)
 	},{	"BEEP_ON_OUTPUT",
 		OPT_BOOL,
 		0,
@@ -116,12 +138,48 @@ struct global_var global_var[] = {
 		opt_set_color,
 		pork_acct_update_blist_color,
 		SET_INT(DEFAULT_COLOR_BLIST_SELECTOR),
+	},{	"CONNECT_TIMEOUT",
+		OPT_INT,
+		0,
+		opt_set_int,
+		NULL,
+		SET_INT(DEFAULT_CONNECT_TIMEOUT)
 	},{	"DOWNLOAD_DIR",
 		OPT_STR,
 		0,
 		opt_set_str,
 		NULL,
 		SET_STR(DEFAULT_DOWNLOAD_DIR),
+	},{	"DUMP_MSGS_TO_STATUS",
+		OPT_BOOL,
+		0,
+		opt_set_bool,
+		NULL,
+		SET_BOOL(DEFAULT_DUMP_MSGS_TO_STATUS),
+	},{	"FORMAT_ACTION_RECV",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_ACTION_RECV),
+	},{	"FORMAT_ACTION_RECV_STATUS",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_ACTION_RECV_STATUS),
+	},{	"FORMAT_ACTION_SEND",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_ACTION_SEND),
+	},{	"FORMAT_ACTION_SEND_STATUS",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_ACTION_SEND_STATUS),
 	},{	"FORMAT_BLIST",
 		OPT_FORMAT,
 		0,
@@ -158,30 +216,54 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_CHAT_IGNORE),
-	},{	"FORMAT_CHAT_INFO",
-		OPT_FORMAT,
-		0,
-		opt_set_format,
-		NULL,
-		SET_STR(DEFAULT_FORMAT_CHAT_INFO),
 	},{	"FORMAT_CHAT_INVITE",
 		OPT_FORMAT,
 		0,
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_CHAT_INVITE),
+	},{	"FORMAT_CHAT_JOIN",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_CHAT_JOIN),
 	},{	"FORMAT_CHAT_KICK",
 		OPT_FORMAT,
 		0,
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_CHAT_KICK),
+	},{	"FORMAT_CHAT_LEAVE",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_CHAT_LEAVE),
+	},{	"FORMAT_CHAT_MODE",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_CHAT_MODE),
+	},{	"FORMAT_CHAT_QUIT",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_CHAT_QUIT),
 	},{	"FORMAT_CHAT_RECV",
 		OPT_FORMAT,
 		0,
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_CHAT_RECV),
+	},{	"FORMAT_CHAT_RECV_ACTION",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_CHAT_RECV_ACTION),
 	},{	"FORMAT_CHAT_RECV_NOTICE",
 		OPT_FORMAT,
 		0,
@@ -194,6 +276,12 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_CHAT_SEND),
+	},{	"FORMAT_CHAT_SEND_ACTION",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_CHAT_SEND_ACTION),
 	},{	"FORMAT_CHAT_SEND_NOTICE",
 		OPT_FORMAT,
 		0,
@@ -212,6 +300,24 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_CHAT_UNIGNORE),
+	},{	"FORMAT_FILE_CANCEL_LOCAL",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_FILE_CANCEL_LOCAL),
+	},{	"FORMAT_FILE_CANCEL_REMOTE",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_FILE_CANCEL_REMOTE),
+	},{	"FORMAT_FILE_LOST",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_FILE_LOST),
 	},{	"FORMAT_FILE_RECV_ACCEPT",
 		OPT_FORMAT,
 		0,
@@ -230,6 +336,12 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_FILE_RECV_COMPLETE),
+	},{	"FORMAT_FILE_RECV_RESUME",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_FILE_RECV_RESUME),
 	},{	"FORMAT_FILE_SEND_ACCEPT",
 		OPT_FORMAT,
 		0,
@@ -248,6 +360,12 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_FILE_SEND_COMPLETE),
+	},{	"FORMAT_FILE_SEND_RESUME",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_FILE_SEND_RESUME),
 	},{	"FORMAT_IM_RECV",
 		OPT_FORMAT,
 		0,
@@ -260,6 +378,12 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_IM_RECV_AUTO),
+	},{	"FORMAT_IM_RECV_STATUS",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_IM_RECV_STATUS),
 	},{	"FORMAT_IM_SEND",
 		OPT_FORMAT,
 		0,
@@ -272,18 +396,36 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_IM_SEND_AUTO),
+	},{	"FORMAT_IM_SEND_STATUS",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_IM_SEND_STATUS),
 	},{	"FORMAT_NOTICE_RECV",
 		OPT_FORMAT,
 		0,
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_NOTICE_RECV),
+	},{	"FORMAT_NOTICE_RECV_STATUS",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_NOTICE_RECV_STATUS),
 	},{	"FORMAT_NOTICE_SEND",
 		OPT_FORMAT,
 		0,
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_NOTICE_SEND),
+	},{	"FORMAT_NOTICE_SEND_STATUS",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_NOTICE_SEND_STATUS),
 	},{	"FORMAT_STATUS",
 		OPT_FORMAT,
 		0,
@@ -296,6 +438,12 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_STATUS_ACTIVITY),
+	},{	"FORMAT_STATUS_CHAT",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_STATUS_CHAT),
 	},{	"FORMAT_STATUS_HELD",
 		OPT_FORMAT,
 		0,
@@ -332,6 +480,48 @@ struct global_var global_var[] = {
 		opt_set_format,
 		NULL,
 		SET_STR(DEFAULT_FORMAT_WARN),
+	},{	"FORMAT_WHOIS_AWAY",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_WHOIS_AWAY),
+	},{ "FORMAT_WHOIS_IDLE",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_WHOIS_IDLE),
+	},{ "FORMAT_WHOIS_MEMBER",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_WHOIS_MEMBER),
+	},{ "FORMAT_WHOIS_NAME",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_WHOIS_NAME),
+	},{ "FORMAT_WHOIS_SIGNON",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_WHOIS_SIGNON),
+	},{	"FORMAT_WHOIS_USERINFO",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_WHOIS_USERINFO),
+	},{	"FORMAT_WHOIS_WARNLEVEL",
+		OPT_FORMAT,
+		0,
+		opt_set_format,
+		NULL,
+		SET_STR(DEFAULT_FORMAT_WHOIS_WARNLEVEL),
 	},{	"HISTORY_LEN",
 		OPT_INT,
 		0,
@@ -356,6 +546,30 @@ struct global_var global_var[] = {
 		opt_set_bool,
 		NULL,
 		SET_BOOL(DEFAULT_LOGIN_ON_STARTUP),
+	},{ "OUTGOING_MSG_FONT",
+		OPT_STR,
+		0,
+		opt_set_str,
+		NULL,
+		SET_STR(DEFAULT_OUTGOING_MSG_FONT),
+	},{ "OUTGOING_MSG_FONT_BGCOLOR",
+		OPT_STR,
+		0,
+		opt_set_str,
+		NULL,
+		SET_STR(DEFAULT_OUTGOING_MSG_FONT_BGCOLOR),
+	},{ "OUTGOING_MSG_FONT_FGCOLOR",
+		OPT_STR,
+		0,
+		opt_set_str,
+		NULL,
+		SET_STR(DEFAULT_OUTGOING_MSG_FONT_FGCOLOR),
+	},{ "OUTGOING_MSG_FONT_SIZE",
+		OPT_STR,
+		0,
+		opt_set_str,
+		NULL,
+		SET_STR(DEFAULT_OUTGOING_MSG_FONT_SIZE),
 	},{	"PORK_DIR",
 		OPT_STR,
 		0,
@@ -374,6 +588,24 @@ struct global_var global_var[] = {
 		opt_set_format,
 		opt_changed_prompt,
 		SET_STR(DEFAULT_PROMPT),
+	},{	"RECONNECT_INTERVAL",
+		OPT_INT,
+		0,
+		opt_set_int,
+		NULL,
+		SET_INT(DEFAULT_RECONNECT_INTERVAL),
+	},{	"RECONNECT_MAX_INTERVAL",
+		OPT_INT,
+		0,
+		opt_set_int,
+		NULL,
+		SET_INT(DEFAULT_RECONNECT_MAX_INTERVAL),
+	},{	"RECONNECT_TRIES",
+		OPT_INT,
+		0,
+		opt_set_int,
+		NULL,
+		SET_INT(DEFAULT_RECONNECT_TRIES),
 	},{	"RECURSIVE_EVENTS",
 		OPT_BOOL,
 		0,
@@ -408,7 +640,7 @@ struct global_var global_var[] = {
 		OPT_INT,
 		0,
 		opt_set_int,
-		NULL,
+		scrollbuf_len_update,
 		SET_INT(DEFAULT_SCROLLBUF_LEN),
 	},{	"SEND_REMOVES_AWAY",
 		OPT_BOOL,
@@ -470,18 +702,6 @@ struct global_var global_var[] = {
 		opt_set_format,
 		pork_acct_update_blist_format,
 		SET_STR(DEFAULT_TEXT_BUDDY_IDLE),
-	},{	"TEXT_CHAT_JOIN",
-		OPT_STR,
-		0,
-		opt_set_format,
-		NULL,
-		SET_STR(DEFAULT_TEXT_CHAT_JOIN),
-	},{	"TEXT_CHAT_LEAVE",
-		OPT_STR,
-		0,
-		opt_set_format,
-		NULL,
-		SET_STR(DEFAULT_TEXT_CHAT_LEAVE),
 	},{	"TEXT_NO_NAME",
 		OPT_STR,
 		0,
@@ -518,6 +738,30 @@ struct global_var global_var[] = {
 		opt_set_bool,
 		NULL,
 		SET_BOOL(DEFAULT_TIMESTAMP),
+	},{	"TRANSFER_PORT_MAX",
+		OPT_INT,
+		0,
+		opt_set_int,
+		NULL,
+		SET_INT(DEFAULT_TRANSFER_PORT_MAX),
+	},{	"TRANSFER_PORT_MIN",
+		OPT_INT,
+		0,
+		opt_set_int,
+		NULL,
+		SET_INT(DEFAULT_TRANSFER_PORT_MIN),
+	},{	"WORDWRAP",
+		OPT_BOOL,
+		0,
+		opt_set_bool,
+		NULL,
+		SET_BOOL(DEFAULT_WORDWRAP),
+	},{	"WORDWRAP_CHAR",
+		OPT_CHAR,
+		0,
+		opt_set_char,
+		NULL,
+		SET_CHAR(DEFAULT_WORDWRAP_CHAR),
 	}
 };
 
@@ -569,7 +813,11 @@ static struct window_var window_var[] = {
 		OPT_BOOL,
 		wopt_set_bool,
 		wopt_changed_timestamp
-	},
+	},{	"WORDWRAP",
+		OPT_BOOL,
+		wopt_set_bool,
+		wopt_changed_wordwrap
+	}
 };
 
 /*
@@ -584,8 +832,10 @@ void wopt_print_var(struct imwindow *imwindow, int i, const char *text) {
 			break;
 
 		case OPT_STR:
-			if (imwindow->opts[i].val.s != NULL) {
-				screen_nocolor_msg("%s %s %s", window_var[i].name,
+			if (imwindow->opts[i].val.s != NULL &&
+				imwindow->opts[i].val.s[0] != '\0')
+			{
+				screen_nocolor_msg("%s %s \"%s\"", window_var[i].name,
 					text, imwindow->opts[i].val.s);
 			} else
 				screen_nocolor_msg("%s is <UNSET>", window_var[i].name);
@@ -597,8 +847,13 @@ void wopt_print_var(struct imwindow *imwindow, int i, const char *text) {
 			break;
 
 		case OPT_CHAR:
-			screen_nocolor_msg("%s %s %c", window_var[i].name, text,
-				imwindow->opts[i].val.c);
+			if (isprint(imwindow->opts[i].val.c)) {
+				screen_nocolor_msg("%s %s '%c'", window_var[i].name, text,
+					imwindow->opts[i].val.c);
+			} else {
+				screen_nocolor_msg("%s %s 0x%x", window_var[i].name, text,
+					imwindow->opts[i].val.c);
+			}
 			break;
 	}
 }
@@ -642,8 +897,9 @@ int wopt_get_val(	struct imwindow *imwindow,
 					char *buf,
 					size_t len)
 {
-	int i = wopt_find(opt_name);
+	int i;
 
+	i = wopt_find(opt_name);
 	if (i == -1)
 		return (-1);
 
@@ -693,8 +949,10 @@ void opt_print_var(int i, const char *text) {
 			break;
 
 		case OPT_STR:
-			if (global_var[i].val.s != NULL) {
-				screen_nocolor_msg("%s %s %s", global_var[i].name,
+			if (global_var[i].val.s != NULL &&
+				global_var[i].val.s[0] != '\0')
+			{
+				screen_nocolor_msg("%s %s \"%s\"", global_var[i].name,
 					text, global_var[i].val.s);
 			} else
 				screen_nocolor_msg("%s is <UNSET>", global_var[i].name);
@@ -707,12 +965,16 @@ void opt_print_var(int i, const char *text) {
 			break;
 
 		case OPT_CHAR:
-			screen_nocolor_msg("%s %s %c", global_var[i].name,
-				text, global_var[i].val.c);
+			if (isprint(global_var[i].val.c)) {
+				screen_nocolor_msg("%s %s '%c'", global_var[i].name,
+					text, global_var[i].val.c);
+			} else {
+				screen_nocolor_msg("%s %s 0x%x", global_var[i].name,
+					text, global_var[i].val.c);
+			}
 			break;
 
-		case OPT_COLOR:
-		{
+		case OPT_COLOR: {
 			char buf[128];
 
 			color_get_str(global_var[i].val.i, buf, sizeof(buf));
@@ -722,6 +984,14 @@ void opt_print_var(int i, const char *text) {
 	}
 }
 
+static void scrollbuf_len_update(void) {
+	u_int32_t scrollbuf_len;
+
+	scrollbuf_len = opt_get_int(OPT_SCROLLBUF_LEN);
+	if (scrollbuf_len < cur_window()->swindow.rows)
+		global_var[OPT_SCROLLBUF_LEN].val.i = cur_window()->swindow.rows;
+}
+
 /*
 ** Print the values of all the global variables.
 */
@@ -737,32 +1007,41 @@ void opt_write(FILE *fp) {
 	size_t i;
 
 	for (i = 0 ; i < array_elem(global_var) ; i++) {
-		fprintf(fp, "set %s ", global_var[i].name);
-
 		switch (global_var[i].type) {
 			case OPT_BOOL:
-				fprintf(fp, "%s\n", (global_var[i].val.b ? "TRUE" : "FALSE"));
+				fprintf(fp, "set %s %s\n", global_var[i].name,
+					(global_var[i].val.b ? "TRUE" : "FALSE"));
 				break;
 
 			case OPT_STR:
-				if (global_var[i].val.s != NULL)
-					fprintf(fp, "%s\n", global_var[i].val.s);
+				if (global_var[i].val.s != NULL &&
+					global_var[i].val.s[0] != '\0')
+				{
+					fprintf(fp, "set %s %s\n", global_var[i].name,
+						global_var[i].val.s);
+				}
 				break;
 
 			case OPT_INT:
-				fprintf(fp, "%d\n", global_var[i].val.i);
+				fprintf(fp, "set %s %d\n", global_var[i].name,
+					global_var[i].val.i);
 				break;
 
 			case OPT_CHAR:
-				fprintf(fp, "%c\n", global_var[i].val.c);
+				if (isprint(global_var[i].val.c)) {
+					fprintf(fp, "set %s %c\n", global_var[i].name,
+						global_var[i].val.c);
+				} else {
+					fprintf(fp, "set %s 0x%x\n", global_var[i].name,
+						global_var[i].val.c);
+				}
 				break;
 
-			case OPT_COLOR:
-			{
+			case OPT_COLOR: {
 				char buf[128];
 
 				color_get_str(global_var[i].val.i, buf, sizeof(buf));
-				fprintf(fp, "%s\n", buf);
+				fprintf(fp, "set %s %s\n", global_var[i].name, buf);
 				break;
 			}
 		}
@@ -825,15 +1104,17 @@ static void opt_changed_prompt(void) {
 }
 
 static void wopt_changed_histlen(struct imwindow *imwindow) {
-	u_int32_t new_len = wopt_get_int(imwindow->opts, WOPT_HISTORY_LEN);
+	u_int32_t new_len;
 
+	new_len = wopt_get_int(imwindow->opts, WOPT_HISTORY_LEN);
 	imwindow->input->history_len = new_len;
 	input_history_prune(imwindow->input);
 }
 
 static void wopt_changed_log(struct imwindow *imwindow) {
-	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_LOG);
+	u_int32_t new_val;
 
+	new_val = wopt_get_bool(imwindow->opts, WOPT_LOG);
 	if (new_val != imwindow->swindow.logged) {
 		if (new_val == 0)
 			swindow_end_log(&imwindow->swindow);
@@ -850,37 +1131,48 @@ static void wopt_changed_logfile(struct imwindow *imwindow) {
 }
 
 static void wopt_changed_priv_input(struct imwindow *imwindow) {
-	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT);
+	u_int32_t new_val;
 
+	new_val = wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT);
 	imwindow_set_priv_input(imwindow, new_val);
 }
 
 static void wopt_changed_scrollbuf_len(struct imwindow *imwindow) {
-	u_int32_t new_len = wopt_get_int(imwindow->opts, WOPT_SCROLLBUF_LEN);
+	u_int32_t new_len;
+
+	new_len = wopt_get_int(imwindow->opts, WOPT_SCROLLBUF_LEN);
+	if (new_len < imwindow->swindow.rows) {
+		imwindow->opts[WOPT_SCROLLBUF_LEN].val.i = imwindow->swindow.rows;
+		imwindow->swindow.scrollbuf_max = imwindow->swindow.rows;
+	} else
+		imwindow->swindow.scrollbuf_max = new_len;
 
-	imwindow->swindow.scrollbuf_max = new_len;
 	swindow_prune(&imwindow->swindow);
 }
 
 static void wopt_changed_scroll_on_output(struct imwindow *imwindow) {
-	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_SCROLL_ON_OUTPUT);
+	u_int32_t new_val;
 
+	new_val = wopt_get_bool(imwindow->opts, WOPT_SCROLL_ON_OUTPUT);
 	imwindow->swindow.scroll_on_output = new_val;
 }
 
 static void wopt_changed_scroll_on_input(struct imwindow *imwindow) {
-	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_SCROLL_ON_INPUT);
+	u_int32_t new_val;
 
+	new_val = wopt_get_bool(imwindow->opts, WOPT_SCROLL_ON_INPUT);
 	imwindow->swindow.scroll_on_input = new_val;
 }
 
 static void wopt_changed_show_blist(struct imwindow *imwindow) {
-	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_SHOW_BLIST);
-	struct pork_acct *acct = imwindow->owner;
+	u_int32_t new_val;
+	struct pork_acct *acct;
 
+	acct = imwindow->owner;
 	if (acct->blist == NULL)
 		return;
 
+	new_val = wopt_get_bool(imwindow->opts, WOPT_SHOW_BLIST);
 	if (new_val == imwindow->blist_visible)
 		return;
 
@@ -895,6 +1187,11 @@ static void wopt_changed_timestamp(struct imwindow *imwindow) {
 		wopt_get_bool(imwindow->opts, WOPT_TIMESTAMP));
 }
 
+static void wopt_changed_wordwrap(struct imwindow *imwindow) {
+	swindow_set_wordwrap(&imwindow->swindow,
+		wopt_get_bool(imwindow->opts, WOPT_WORDWRAP));
+}
+
 void wopt_init(struct imwindow *imwindow, const char *target) {
 	char nnick[NUSER_LEN];
 	char buf[NUSER_LEN + 256];
@@ -913,6 +1210,7 @@ void wopt_init(struct imwindow *imwindow, const char *target) {
 	wopt[WOPT_SCROLLBUF_LEN].val.i = opt_get_int(OPT_SCROLLBUF_LEN);
 	wopt[WOPT_SHOW_BLIST].val.b = opt_get_bool(OPT_SHOW_BLIST);
 	wopt[WOPT_TIMESTAMP].val.b = opt_get_bool(OPT_TIMESTAMP);
+	wopt[WOPT_WORDWRAP].val.b = opt_get_bool(OPT_WORDWRAP);
 
 	normalize(nnick, target, sizeof(nnick));
 	while ((p = strchr(nnick, '/')) != NULL)
@@ -976,7 +1274,15 @@ static int opt_set_char(u_int32_t opt, char *args) {
 	if (args == NULL || *args == '\0')
 		return (-1);
 
-	global_var[opt].val.c = *args;
+	if (!strncasecmp(args, "0x", 2)) {
+		u_int32_t temp;
+
+		if (str_to_uint(args, &temp) == -1 || temp > 0xff)
+			global_var[opt].val.c = *args;
+		else
+			global_var[opt].val.c = temp;
+	} else
+		global_var[opt].val.c = *args;
 
 	if (global_var[opt].updated != NULL)
 		global_var[opt].updated();
@@ -1071,7 +1377,15 @@ static int wopt_set_char(struct imwindow *imwindow, u_int32_t opt, char *args) {
 	if (args == NULL || *args == '\0')
 		return (-1);
 
-	imwindow->opts[opt].val.c = *args;
+	if (!strncasecmp(args, "0x", 2)) {
+		u_int32_t temp;
+
+		if (str_to_uint(args, &temp) == -1 || temp > 0xff)
+			imwindow->opts[opt].val.c = *args;
+		else
+			imwindow->opts[opt].val.c = temp;
+	} else
+		imwindow->opts[opt].val.c = *args;
 
 	if (window_var[opt].updated != NULL)
 		window_var[opt].updated(imwindow);
