@@ -1,6 +1,6 @@
 /*
 ** pork_swindow.c - an ncurses scrolling window widget
-** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
+** Copyright (C) 2002-2004 Ryan McCabe <ryan@numb.org>
 **
 ** I've tried to keep this code independent from the rest of the program
 ** so that it might be useful outside of this particular program.
@@ -34,6 +34,64 @@
 
 static void swindow_scroll(struct swindow *swindow, int n);
 
+
+static u_int32_t imsg_wordwrapped_lines(struct swindow *swindow,
+										struct imsg *imsg)
+{
+	u_int32_t len = MSG_LEN(swindow, imsg);
+	chtype *ch = MSG_TEXT(swindow, imsg);
+	chtype *end = &ch[len - 1];
+	u_int32_t lines = 0;
+	int add = 0;
+	chtype cont_char = opt_get_char(OPT_WORDWRAP_CHAR);
+
+	if (len <= swindow->cols)
+		return (1);
+
+	if (swindow->cols == 1)
+		return (len);
+
+	while (*ch != 0) {
+		chtype *p;
+
+		p = &ch[swindow->cols - 1 - add];
+		if (p >= end) {
+			++lines;
+			break;
+		}
+
+		/*
+		** If the last character on current line or the first character
+		** on the next line is a space, there's no need to break the line
+		*/
+		if (chtype_get(p[0]) != ' ' && chtype_get(p[1]) != ' ') {
+			chtype *temp = p;
+
+			/* Find the last space, if any */
+			do {
+				temp--;
+				if (temp <= ch) {
+					/*
+					** No spaces on the line, just split it at the
+					** last column
+					*/
+					goto out;
+				}
+			} while (chtype_get(*temp) != ' ');
+			/* Found a space, split the line after the space. */
+			p = temp;
+		}
+out:
+		if (cont_char != 0)
+			add = 1;
+
+		ch = &p[1];
+		lines++;
+	}
+
+	return (lines);
+}
+
 /*
 ** Compute the number of rows the message will occupy
 ** on the screen.
@@ -41,18 +99,14 @@ static void swindow_scroll(struct swindow *swindow, int n);
 
 static inline u_int32_t imsg_lines(struct swindow *swindow, struct imsg *imsg) {
 	u_int32_t len = MSG_LEN(swindow, imsg);
-	u_int32_t lines;
 
-	/*
-	** With timestamps turned off, blank lines have a length of 0, but
-	** we still need one line to display them.
-	*/
-
-	if (len == 0)
+	if (len <= swindow->cols)
 		return (1);
 
-	lines = (len / swindow->cols) + (len % swindow->cols != 0);
-	return (lines);
+	if (!swindow->wordwrap)
+		return ((len / swindow->cols) + (len % swindow->cols != 0));
+
+	return (imsg_wordwrapped_lines(swindow, imsg));
 }
 
 /*
@@ -72,11 +126,11 @@ static struct imsg *imsg_new(	struct swindow *swindow,
 	chtype *ch;
 	chtype *c;
 
-	if (timestamp == 1) {
+	if (timestamp) {
 		time_t cur_time;
 		struct tm *tm;
 
-		cur_time = time(NULL);
+		time(&cur_time);
 		tm = localtime(&cur_time);
 		snprintf(timebuf, sizeof(timebuf), "[%02d:%02d] ",
 			tm->tm_hour, tm->tm_min);
@@ -98,6 +152,7 @@ static struct imsg *imsg_new(	struct swindow *swindow,
 
 	imsg = xmalloc(sizeof(*imsg));
 	imsg->text = msg_text;
+	imsg->serial = swindow->serial++;
 	imsg->timestamp = timestamp;
 	imsg->len = len - 1;
 	imsg->lines = imsg_lines(swindow, imsg);
@@ -142,6 +197,99 @@ static chtype *imsg_partial(struct swindow *swindow,
 	return (&nul_ch);
 }
 
+static int swindow_print_msg_wr(struct swindow *swindow,
+								struct imsg *imsg,
+								u_int32_t y,
+								u_int32_t x,
+								u_int32_t firstline,
+								u_int32_t lastline)
+{
+	u_int32_t len = MSG_LEN(swindow, imsg);
+	chtype *ch = MSG_TEXT(swindow, imsg);
+	u_int32_t lines = 0;
+	u_int32_t lines_printed = 0;
+	chtype *end = &ch[len - 1];
+	int add = 0;
+	chtype cont_char = (chtype) opt_get_char(OPT_WORDWRAP_CHAR);
+
+	if (swindow->cols == 1)
+		cont_char = 0;
+
+	while (*ch != 0) {
+		chtype *p;
+
+		p = &ch[swindow->cols - 1 - add];
+		if (p >= end) {
+			++lines;
+			if (lines <= lastline) {
+				mvwputstr(swindow->win, y + lines_printed, add, ch);
+				++lines_printed;
+			}
+			break;
+		}
+
+		if (chtype_get(p[0]) != ' ' && chtype_get(p[1]) != ' ') {
+			chtype *temp = p;
+
+			do {
+				temp--;
+				if (temp <= ch)
+					goto out;
+			} while (chtype_get(*temp) != ' ');
+			p = temp;
+		}
+out:
+		++lines;
+		if (lines >= firstline) {
+			if (lines > lastline)
+				break;
+
+			mvwputnstr(swindow->win, y + lines_printed, add, ch, p - ch + 1);
+			++lines_printed;
+		}
+
+		if (cont_char != 0) {
+			mvwaddch(swindow->win, y + lines_printed, 0, cont_char);
+			add = 1;
+		}
+		ch = &p[1];
+	}
+
+	return (lines_printed);
+}
+
+static int swindow_print_msg(	struct swindow *swindow,
+								struct imsg *imsg,
+								u_int32_t y,
+								u_int32_t x,
+								u_int32_t firstline,
+								u_int32_t lastline)
+{
+	chtype *msg;
+
+	if (swindow->wordwrap)
+		return (swindow_print_msg_wr(swindow, imsg, y, x, firstline, lastline));
+
+	if (lastline < firstline)
+		return (-1);
+
+	msg = imsg_partial(swindow, imsg, firstline);
+	if (lastline >= imsg->lines)
+		mvwputstr(swindow->win, y, x, msg);
+	else {
+		u_int32_t max_len;
+
+		if (lastline == firstline)
+			max_len = swindow->cols;
+		else
+			max_len = (lastline - firstline) * swindow->cols;
+
+		mvwputnstr(swindow->win, y, x, msg, max_len);
+	}
+
+	return (0);
+}
+
 /*
 ** Initialize a new window.
 */
@@ -166,9 +314,13 @@ int swindow_init(	struct swindow *swindow,
 	swindow->scroll_on_input = wopt_get_bool(wopt, WOPT_SCROLL_ON_INPUT);
 	swindow->scroll_on_output = wopt_get_bool(wopt, WOPT_SCROLL_ON_OUTPUT);
 	swindow->timestamp = wopt_get_bool(wopt, WOPT_TIMESTAMP);
+	swindow->wordwrap = wopt_get_bool(wopt, WOPT_WORDWRAP);
 	swindow->logged = wopt_get_bool(wopt, WOPT_LOG);
 	swindow->logfile = wopt_get_str(wopt, WOPT_LOGFILE);
 
+	if (swindow->scrollbuf_max < rows)
+		swindow->scrollbuf_max = rows;
+
 	if (swindow->logged)
 		swindow_set_log(swindow);
 
@@ -193,25 +345,23 @@ static inline void swindow_scroll(struct swindow *swindow, int n) {
 ** Prune the scroll buffer so that the number of total
 ** messages is not greater than swindow->scrollbuf_max.
 **
-** This should not be called if the window is scrolled up.
 */
 
 void swindow_prune(struct swindow *swindow) {
-	u_int32_t num = swindow->scrollbuf_len - swindow->scrollbuf_max;
+	int num = swindow->scrollbuf_len - swindow->scrollbuf_max;
 	dlist_t *cur = swindow->scrollbuf_end;
+	u_int32_t serial_top;
+	u_int32_t serial_bot;
 
-	if (swindow->bottom_blank > 0)
-		return;
+	serial_top = ((struct imsg *) swindow->scrollbuf_top->data)->serial;
+	serial_bot = ((struct imsg *) swindow->scrollbuf_bot->data)->serial;
 
 	while (cur != NULL && num > 0) {
 		struct imsg *imsg = cur->data;
 		dlist_t *next = cur->prev;
 
-		/*
-		** Don't prune any messages are still on the screen.
-		*/
-
-		if (swindow->scrollbuf_lines - imsg->lines < swindow->rows)
+		/* Don't prune anything that's still on the screen */
+		if (imsg->serial >= serial_top && imsg->serial <= serial_bot)
 			break;
 
 		swindow->scrollbuf_lines -= imsg->lines;
@@ -278,13 +428,10 @@ static void swindow_recalculate(struct swindow *swindow,
 	struct imsg *imsg_top;
 	u_int32_t old_top;
 
-	/*
-	** If nothing's been printed to this
-	** window, there's no work to do here.
-	*/
-
-	if (swindow->scrollbuf_top == NULL)
+	if (swindow->scrollbuf_top == NULL) {
+		swindow->bottom_blank = swindow->rows;
 		return;
+	}
 
 	imsg_top = swindow->scrollbuf_top->data;
 	old_top = imsg_top->lines;
@@ -333,6 +480,7 @@ void swindow_resize(struct swindow *swindow,
 {
 	u_int32_t old_rows = swindow->rows;
 	u_int32_t old_cols = swindow->cols;
+	int ret;
 
 	/*
 	** This seems to be necessary to prevent some funny stuff
@@ -342,11 +490,16 @@ void swindow_resize(struct swindow *swindow,
 	wclear(swindow->win);
 	wnoutrefresh(swindow->win);
 
-	wresize(swindow->win, new_row, new_col);
+	ret = wresize(swindow->win, new_row, new_col);
+	if (ret == -1)
+		debug("wresize: %d (%u,%u)", ret, new_row, new_col);
 
 	swindow->rows = new_row;
 	swindow->cols = new_col;
 
+	if (swindow->scrollbuf_max < swindow->rows)
+		swindow->scrollbuf_max = swindow->rows;
+
 	swindow_recalculate(swindow, old_rows, old_cols);
 	swindow_redraw(swindow);
 	swindow->dirty = 1;
@@ -370,71 +523,56 @@ inline int swindow_refresh(struct swindow *swindow) {
 
 void swindow_redraw(struct swindow *swindow) {
 	dlist_t *cur = swindow->scrollbuf_top;
-	dlist_t *save = NULL;
-	u_int32_t lines = 0;
 	u_int32_t curs_pos = 0;
-	u_int32_t top_hidden = swindow->top_hidden;
-
-	while (cur != NULL && lines < swindow->rows) {
-		struct imsg *imsg = cur->data;
-
-		lines += imsg->lines;
-		if (top_hidden != 0) {
-			lines -= top_hidden;
-			top_hidden = 0;
-		}
-
-		if (lines >= swindow->rows)
-			break;
-
-		save = cur;
-		cur = cur->prev;
-	}
-
-	if (cur != NULL) {
-		swindow->scrollbuf_bot = cur;
-		swindow->bottom_hidden = lines - swindow->rows;
-		swindow->bottom_blank = 0;
-	} else {
-		swindow->scrollbuf_bot = save;
-		swindow->bottom_hidden = 0;
-		swindow->held = 0;
-		swindow->bottom_blank = swindow->rows - lines;
-	}
-
-	cur = swindow->scrollbuf_top;
 
+	if (cur == NULL)
+		return;
 	/*
 	** If part of the top message is scrolled off
 	** the top, print the visible part and advance
 	** the pointer to the next message.
 	*/
-
 	if (swindow->top_hidden != 0) {
 		struct imsg *imsg = cur->data;
-		chtype *partial;
-
-		partial = imsg_partial(swindow, imsg, swindow->top_hidden + 1);
-		mvwputstr(swindow->win, 0, 0, partial);
 
+		swindow_print_msg(swindow, imsg, 0, 0, swindow->top_hidden + 1, -1);
 		curs_pos += imsg->lines - swindow->top_hidden;
+		if (curs_pos > swindow->rows) {
+			swindow->bottom_blank = 0;
+			swindow->bottom_hidden = curs_pos - swindow->rows;
+			swindow->scrollbuf_bot = cur;
+		}
+
 		cur = cur->prev;
 	}
 
-	while (cur != NULL && curs_pos <= swindow->rows - 1) {
+	while (cur != NULL && curs_pos < swindow->rows) {
 		struct imsg *imsg = cur->data;
 
 		if (curs_pos + imsg->lines > swindow->rows) {
-			u_int32_t max_len = (swindow->rows - curs_pos) * swindow->cols;
-
-			mvwputnstr(swindow->win, curs_pos, 0,
-				MSG_TEXT(swindow, imsg), max_len);
+			swindow->scrollbuf_bot = cur;
+			swindow->bottom_blank = 0;
+			swindow->bottom_hidden = imsg->lines - (swindow->rows - curs_pos);
 
+			swindow_print_msg(swindow, imsg, curs_pos, 0,
+				1, swindow->rows - curs_pos);
 			break;
 		}
 
-		mvwputstr(swindow->win, curs_pos, 0, MSG_TEXT(swindow, imsg));
+		swindow_print_msg(swindow, imsg, curs_pos, 0, 1, -1);
 		curs_pos += imsg->lines;
+
+		if (cur->prev == NULL) {
+			swindow->scrollbuf_bot = cur;
+			swindow->bottom_blank = swindow->rows - curs_pos;
+			swindow->bottom_hidden = 0;
+			swindow->held = 0;
+		} else if (curs_pos == swindow->rows) {
+			swindow->scrollbuf_bot = cur;
+			swindow->bottom_hidden = 0;
+			swindow->bottom_blank = 0;
+		}
+
 		cur = cur->prev;
 	}
 
@@ -451,6 +589,7 @@ void swindow_redraw(struct swindow *swindow) {
 int swindow_add(struct swindow *swindow, chtype *msg, u_int32_t timestamp) {
 	struct imsg *imsg;
 	chtype *msg_text;
+	u_int32_t msg_line_start = 1;
 	dlist_t *old_head = swindow->scrollbuf;
 	int y_pos;
 
@@ -502,13 +641,6 @@ int swindow_add(struct swindow *swindow, chtype *msg, u_int32_t timestamp) {
 		swindow->scrollbuf_top = swindow->scrollbuf;
 	}
 
-	/*
-	** If there's a maximum scroll buffer length, enforce it.
-	*/
-
-	if (swindow->scrollbuf_len > swindow->scrollbuf_max)
-		swindow_prune(swindow);
-
 	/*
 	** If the window is scrolled back, but the scroll
 	** on output flag is not set, there's nothing left
@@ -548,18 +680,25 @@ int swindow_add(struct swindow *swindow, chtype *msg, u_int32_t timestamp) {
 		y_pos = swindow->rows - imsg->lines;
 		if (y_pos < 0) {
 			y_pos = 0;
-			msg_text = imsg_partial(swindow, imsg, swindow->top_hidden + 1);
+			msg_line_start = imsg->lines - swindow->rows + 1;
 		}
 
 		swindow_scroll(swindow, evict);
 	}
 
-	mvwputstr(swindow->win, y_pos, 0, msg_text);
+	swindow_print_msg(swindow, imsg, y_pos, 0, msg_line_start, -1);
 	swindow->dirty = 1;
 
 	if (!swindow->visible)
 		swindow->activity = 1;
 
+	/*
+	** If there's a maximum scroll buffer length, enforce it.
+	*/
+
+	if (swindow->scrollbuf_len > swindow->scrollbuf_max)
+		swindow_prune(swindow);
+
 	return (0);
 }
 
@@ -568,7 +707,7 @@ int swindow_add(struct swindow *swindow, chtype *msg, u_int32_t timestamp) {
 ** a message to a window that was typed by the user.
 */
 
-inline int swindow_input(struct swindow *swindow, chtype *msg) {
+inline int swindow_input(struct swindow *swindow) {
 	/*
 	** If the window is scrolled up, and the scroll on
 	** input flag is set, scroll to the bottom of the
@@ -582,13 +721,16 @@ inline int swindow_input(struct swindow *swindow, chtype *msg) {
 		swindow_scroll_to_end(swindow);
 	}
 
-	return (swindow_add(swindow, msg, 1));
+	return (0);
 }
 
+/*
+** Scroll up one line.
+*/
+
 u_int32_t swindow_scroll_up(struct swindow *swindow) {
 	struct imsg *imsg;
 	u_int32_t top_line;
-	chtype *partial;
 
 	/*
 	** If it's already at the top, it can't scroll anymore.
@@ -639,9 +781,7 @@ u_int32_t swindow_scroll_up(struct swindow *swindow) {
 	}
 
 	top_line = swindow->top_hidden + 1;
-	partial = imsg_partial(swindow, imsg, top_line);
-	mvwputnstr(swindow->win, 0, 0, partial, swindow->cols);
-
+	swindow_print_msg(swindow, imsg, 0, 0, top_line, top_line);
 	return (1);
 }
 
@@ -651,7 +791,6 @@ u_int32_t swindow_scroll_up(struct swindow *swindow) {
 
 u_int32_t swindow_scroll_down(struct swindow *swindow) {
 	struct imsg *imsg;
-	chtype *partial;
 	u_int32_t bottom_line;
 
 	/*
@@ -699,8 +838,9 @@ u_int32_t swindow_scroll_down(struct swindow *swindow) {
 	}
 
 	bottom_line = imsg->lines - swindow->bottom_hidden;
-	partial = imsg_partial(swindow, imsg, bottom_line);
-	mvwputnstr(swindow->win, swindow->rows - 1, 0, partial, swindow->cols);
+
+	swindow_print_msg(swindow, imsg, swindow->rows - 1, 0,
+		bottom_line, bottom_line);
 
 	if (swindow->held > 0)
 		swindow->held--;
@@ -713,8 +853,41 @@ u_int32_t swindow_scroll_down(struct swindow *swindow) {
 */
 
 void swindow_scroll_to_end(struct swindow *swindow) {
-	while (swindow_scroll_down(swindow) != 0)
-		;
+	dlist_t *cur;
+	u_int32_t lines = 0;
+
+	if (swindow->scrollbuf == NULL)
+		return;
+
+	/* Avoid a redraw if it's already at the bottom */
+	if (swindow->scrollbuf_bot == swindow->scrollbuf &&
+		swindow->bottom_hidden == 0)
+	{
+		return;
+	}
+
+	cur = swindow->scrollbuf;
+	do {
+		struct imsg *imsg = cur->data;
+
+		lines += imsg->lines;
+		if (lines >= swindow->rows) {
+			swindow->scrollbuf_top = cur;
+			swindow->top_hidden = lines - swindow->rows;
+			break;
+		}
+
+		if (cur->next == NULL) {
+			swindow->scrollbuf_top = cur;
+			swindow->top_hidden = 0;
+			break;
+		}
+		cur = cur->next;
+	} while (1);
+
+	swindow->held = 0;
+	wclear(swindow->win);
+	swindow_redraw(swindow);
 }
 
 /*
@@ -722,10 +895,29 @@ void swindow_scroll_to_end(struct swindow *swindow) {
 */
 
 void swindow_scroll_to_start(struct swindow *swindow) {
-	while (swindow_scroll_up(swindow) != 0)
-		;
+	if (swindow->scrollbuf == NULL)
+		return;
+
+	/* Avoid a redraw if it's already at the top */
+	if (swindow->scrollbuf_top == swindow->scrollbuf_end &&
+		swindow->top_hidden == 0)
+	{
+		return;
+	}
+
+	swindow->top_hidden = 0;
+	swindow->scrollbuf_top = swindow->scrollbuf_end;
+
+	wclear(swindow->win);
+	swindow_redraw(swindow);
 }
 
+/*
+** This should probably be rewritten to adjust the top (and top_hidden)
+** scrollbuf variables and redrawing instead of scrolling up/down one
+** line at a time if "lines" is more than a few lines..
+*/
+
 void swindow_scroll_by(struct swindow *swindow, int lines) {
 	if (lines < 0) {
 		int i;
@@ -758,6 +950,7 @@ int swindow_print_matching(	struct swindow *swindow,
 	int cflags = REG_EXTENDED;
 	regex_t preg;
 	dlist_t *cur;
+	dlist_t *match_list = NULL;
 
 	if (regex == NULL)
 		return (-1);
@@ -776,21 +969,30 @@ int swindow_print_matching(	struct swindow *swindow,
 		char *buf;
 
 		buf = cstr_to_plaintext(MSG_TEXT(swindow, imsg), MSG_LEN(swindow, imsg));
-		if (buf == NULL)
-			continue;
-
-		if (regexec(&preg, buf, 0, NULL, 0) == 0) {
-			size_t len = MSG_LEN(swindow, imsg) + 1;
-			chtype ch[len];
-
-			memcpy(ch, MSG_TEXT(swindow, imsg), len * sizeof(chtype));
-			swindow_add(swindow, ch, 0);
+		if (buf != NULL) {
+			if (regexec(&preg, buf, 0, NULL, 0) == 0)
+				match_list = dlist_add_head(match_list, imsg);
+			free(buf);
 		}
-
-		free(buf);
 	}
 
 	regfree(&preg);
+
+	/*
+	** Better to compile a list of matches and print them after scanning
+	** the whole buffer. If we started scanning at the oldest message and
+	** printed matches as we traversed the scrollbuffer list, bad interactions
+	** with swindow_prune() could occur.
+	*/
+	cur = match_list;
+	while (cur != NULL) {
+		dlist_t *next = cur->next;
+
+		swindow_add(swindow, MSG_TEXT(swindow, (struct imsg *) cur->data), 0);
+		free(cur);
+		cur = next;
+	}
+
 	return (0);
 }
 
@@ -812,6 +1014,54 @@ inline void swindow_set_timestamp(struct swindow *swindow, u_int32_t value) {
 	swindow_redraw(swindow);
 }
 
+inline void swindow_set_wordwrap(struct swindow *swindow, u_int32_t value) {
+	/*
+	** Allow for updating after the continued char changed but the
+	** wordwrap enabled setting didn't.
+	*/
+	swindow->wordwrap = value;
+	swindow_recalculate(swindow, swindow->rows, swindow->cols);
+	wclear(swindow->win);
+	swindow_redraw(swindow);
+}
+
+/*
+** dump everything in the window to the specified file.
+*/
+
+int swindow_dump_buffer(struct swindow *swindow, char *file) {
+	int fd;
+	dlist_t *cur;
+	struct iovec wvec[2];
+
+	if (swindow->scrollbuf_end == NULL)
+		return (-1);
+
+	fd = open(file, O_WRONLY | O_CREAT | O_APPEND, 0600);
+	if (fd == -1)
+		return (-1);
+
+	wvec[1].iov_base = "\n";
+	wvec[1].iov_len = 1;
+
+	for (cur = swindow->scrollbuf_end ; cur != NULL ; cur = cur->prev) {
+		struct imsg *imsg = cur->data;
+
+		wvec[0].iov_base = cstr_to_plaintext(imsg->text, imsg->len);
+		wvec[0].iov_len = imsg->len;
+
+		if (writev(fd, wvec, 2) != (int) imsg->len + 1) {
+			screen_err_msg("Error writing buffer to %s: %s",
+				file, strerror(errno));
+		}
+
+		free(wvec[0].iov_base);
+	}
+
+	close(fd);
+	return (0);
+}
+
 /*
 ** Set the window to log to "logfile"
 */
@@ -941,12 +1191,15 @@ void swindow_erase(struct swindow *swindow) {
 	swindow->scrollbuf = NULL;
 	swindow->scrollbuf_top = NULL;
 	swindow->scrollbuf_bot = NULL;
+	swindow->scrollbuf_end = NULL;
 	swindow->top_hidden = 0;
 	swindow->bottom_hidden = 0;
 	swindow->scrollbuf_len = 0;
-	swindow->bottom_blank = swindow->rows;
 	swindow->scrollbuf_lines = 0;
 	swindow->held = 0;
+	swindow->serial = 0;
+	swindow->activity = 0;
+	swindow->bottom_blank = swindow->rows;
 	swindow->dirty = 1;
 
 	wclear(swindow->win);
